# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""This module contains all the functions required to save the decomposed result object in the database."""

import os
from datetime import datetime as dt
from datetime import timezone
from pathlib import Path
from typing import Any

import cloudpickle
import networkx as nx
from sqlalchemy import update
from sqlalchemy.orm import Session

from covalent._data_store.datastore import DataStore, workflow_db
from covalent._data_store.models import Electron, ElectronDependency, Lattice

from .._shared_files import logger
from .._shared_files.defaults import (
    arg_prefix,
    attr_prefix,
    electron_dict_prefix,
    electron_list_prefix,
    generator_prefix,
    parameter_prefix,
    prefix_separator,
    sublattice_prefix,
    subscript_prefix,
)

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class MissingLatticeRecordError(Exception):
    pass


class MissingElectronRecordError(Exception):
    pass


class InvalidFileExtension(Exception):
    pass


def update_lattice_completed_electron_num(dispatch_id: str) -> None:
    """Update the number of completed electrons by one corresponding to a lattice."""

    with workflow_db.session() as session:
        session.query(Lattice).filter_by(dispatch_id=dispatch_id).update(
            {
                "completed_electron_num": Lattice.completed_electron_num + 1,
                "updated_at": dt.now(timezone.utc),
            }
        )
        session.commit()


def insert_lattices_data(
    dispatch_id: str,
    name: str,
    electron_num: int,
    completed_electron_num: int,
    status: str,
    storage_type: str,
    storage_path: str,
    function_filename: str,
    function_string_filename: str,
    executor: str,
    executor_data_filename: str,
    workflow_executor: str,
    workflow_executor_data_filename: str,
    error_filename: str,
    inputs_filename: str,
    named_args_filename: str,
    named_kwargs_filename: str,
    results_filename: str,
    transport_graph_filename: str,
    created_at: dt,
    updated_at: dt,
    started_at: dt,
    completed_at: dt,
) -> int:
    """This funtion writes the lattice data / metadata to the Lattices table in the DB."""

    lattice_row = Lattice(
        dispatch_id=dispatch_id,
        name=name,
        status=status,
        electron_num=electron_num,
        completed_electron_num=completed_electron_num,
        storage_type=storage_type,
        storage_path=storage_path,
        function_filename=function_filename,
        function_string_filename=function_string_filename,
        executor=executor,
        executor_data_filename=executor_data_filename,
        workflow_executor=workflow_executor,
        workflow_executor_data_filename=workflow_executor_data_filename,
        error_filename=error_filename,
        inputs_filename=inputs_filename,
        named_args_filename=named_args_filename,
        named_kwargs_filename=named_kwargs_filename,
        results_filename=results_filename,
        transport_graph_filename=transport_graph_filename,
        is_active=True,
        created_at=created_at,
        updated_at=updated_at,
        started_at=started_at,
        completed_at=completed_at,
    )

    with workflow_db.session() as session:
        session.add(lattice_row)
        lattice_id = lattice_row.id
        session.commit()

    app_log.debug(f"returning lattice id {lattice_id}")
    return lattice_id


def insert_electrons_data(
    parent_dispatch_id: str,
    transport_graph_node_id: int,
    type: str,
    name: str,
    status: str,
    storage_type: str,
    storage_path: str,
    function_filename: str,
    function_string_filename: str,
    executor: str,
    executor_data_filename: str,
    results_filename: str,
    value_filename: str,
    stdout_filename: str,
    stderr_filename: str,
    info_filename: str,
    deps_filename: str,
    call_before_filename: str,
    call_after_filename: str,
    created_at: dt,
    updated_at: dt,
    started_at: dt,
    completed_at: dt,
) -> id:
    """This function writes the transport graph node data to the Electrons table in the DB."""

    # Check that the foreign key corresponding to this table exists
    app_log.debug("insert_electrons_data")
    with workflow_db.session() as session:
        row = session.query(Lattice).where(Lattice.dispatch_id == parent_dispatch_id).all()
        if len(row) == 0:
            raise MissingLatticeRecordError

        parent_lattice_id = row[0].id
        app_log.debug(parent_lattice_id)

        electron_row = Electron(
            parent_lattice_id=parent_lattice_id,
            transport_graph_node_id=transport_graph_node_id,
            type=type,
            name=name,
            status=status,
            storage_type=storage_type,
            storage_path=storage_path,
            function_filename=function_filename,
            function_string_filename=function_string_filename,
            executor=executor,
            executor_data_filename=executor_data_filename,
            results_filename=results_filename,
            value_filename=value_filename,
            stdout_filename=stdout_filename,
            stderr_filename=stderr_filename,
            info_filename=info_filename,
            deps_filename=deps_filename,
            call_before_filename=call_before_filename,
            call_after_filename=call_after_filename,
            is_active=True,
            created_at=created_at,
            updated_at=updated_at,
            started_at=started_at,
            completed_at=completed_at,
        )

    with workflow_db.session() as session:
        session.add(electron_row)
        electron_id = electron_row.id
        session.commit()

    return electron_id


def insert_electron_dependency_data(dispatch_id: str, lattice: "Lattice"):
    """Extract electron dependencies from the lattice transport graph and add them to the DB."""

    # TODO - Update how we access the transport graph edges directly in favor of using some interface provied by the TransportGraph class.
    node_links = nx.readwrite.node_link_data(lattice.transport_graph._graph)["links"]

    electron_dependency_ids = []
    with workflow_db.session() as session:
        for edge_data in node_links:
            electron_id = (
                session.query(Lattice, Electron)
                .filter(Lattice.id == Electron.parent_lattice_id)
                .filter(Lattice.dispatch_id == dispatch_id)
                .filter(Electron.transport_graph_node_id == edge_data["target"])
                .first()
                .Electron.id
            )
            parent_electron_id = (
                session.query(Lattice, Electron)
                .filter(Lattice.id == Electron.parent_lattice_id)
                .filter(Lattice.dispatch_id == dispatch_id)
                .filter(Electron.transport_graph_node_id == edge_data["source"])
                .first()
                .Electron.id
            )

            electron_dependency_row = ElectronDependency(
                electron_id=electron_id,
                parent_electron_id=parent_electron_id,
                edge_name=edge_data["edge_name"],
                parameter_type=edge_data["param_type"] if "param_type" in edge_data else None,
                arg_index=edge_data["arg_index"] if "arg_index" in edge_data else None,
                is_active=True,
                created_at=dt.now(timezone.utc),
                updated_at=dt.now(timezone.utc),
            )

            session.add(electron_dependency_row)
            electron_dependency_ids.append(electron_dependency_row.id)
        session.commit()

    return electron_dependency_ids


def update_lattices_data(dispatch_id: str, **kwargs) -> None:
    """This function updates the lattices record."""

    with workflow_db.session() as session:
        valid_update = session.query(Lattice).where(Lattice.dispatch_id == dispatch_id).first()

        if not valid_update:
            raise MissingLatticeRecordError

        for attr, value in kwargs.items():
            if value:
                setattr(valid_update, attr, value)

        session.add(valid_update)
        session.commit()


def update_electrons_data(
    parent_dispatch_id: str,
    transport_graph_node_id: int,
    status: str,
    started_at: dt,
    updated_at: dt,
    completed_at: dt,
) -> None:
    """This function updates the electrons record."""

    with workflow_db.session() as session:
        parent_lattice_id = (
            session.query(Lattice).where(Lattice.dispatch_id == parent_dispatch_id).all()[0].id
        )
        valid_update = (
            session.query(Electron)
            .where(
                Electron.parent_lattice_id == parent_lattice_id,
                Electron.transport_graph_node_id == transport_graph_node_id,
            )
            .first()
            is not None
        )
        if not valid_update:
            raise MissingElectronRecordError

        session.execute(
            update(Electron)
            .where(
                Electron.parent_lattice_id == parent_lattice_id,
                Electron.transport_graph_node_id == transport_graph_node_id,
            )
            .values(
                status=status,
                started_at=started_at,
                updated_at=updated_at,
                completed_at=completed_at,
            )
        )
        session.commit()


def get_electron_type(node_name: str) -> str:
    """Get the electron type (to be written to DB) given the electron node data."""

    if node_name.startswith(arg_prefix):
        return arg_prefix.strip(prefix_separator)

    elif node_name.startswith(attr_prefix):
        return attr_prefix.strip(prefix_separator)

    elif node_name.startswith(electron_dict_prefix):
        return electron_dict_prefix.strip(prefix_separator)

    elif node_name.startswith(electron_list_prefix):
        return electron_list_prefix.strip(prefix_separator)

    elif node_name.startswith(generator_prefix):
        return generator_prefix.strip(prefix_separator)

    elif node_name.startswith(parameter_prefix):
        return parameter_prefix.strip(prefix_separator)

    elif node_name.startswith(sublattice_prefix):
        return sublattice_prefix.strip(prefix_separator)

    elif node_name.startswith(subscript_prefix):
        return subscript_prefix.strip(prefix_separator)

    else:
        return "function"


def write_sublattice_electron_id(
    parent_dispatch_id: str, sublattice_node_id: int, sublattice_dispatch_id: str
) -> None:
    """Function to attach the electron id of a sublattice in the lattice record."""

    with workflow_db.session() as session:
        sublattice_electron_id = (
            session.query(Lattice, Electron)
            .where(
                Lattice.id == Electron.parent_lattice_id,
                Electron.transport_graph_node_id == sublattice_node_id,
                Lattice.dispatch_id == parent_dispatch_id,
            )
            .first()
            .Electron.id
        )
        session.execute(
            update(Lattice)
            .where(Lattice.dispatch_id == sublattice_dispatch_id)
            .values(electron_id=sublattice_electron_id, updated_at=dt.now(timezone.utc))
        )
        session.commit()


def write_lattice_error(dispatch_id: str, error: str):
    with workflow_db.session() as session:
        valid_update = session.query(Lattice).where(Lattice.dispatch_id == dispatch_id).first()

        if not valid_update:
            raise MissingLatticeRecordError

        with open(os.path.join(valid_update.storage_path, valid_update.error_filename), "w") as f:
            f.write(error)


def store_file(storage_path: str, filename: str, data: Any = None) -> None:
    """This function writes data corresponding to the filepaths in the DB."""

    if filename.endswith(".pkl"):
        with open(Path(storage_path) / filename, "wb") as f:
            cloudpickle.dump(data, f)

    elif filename.endswith(".log") or filename.endswith(".txt"):
        if data is None:
            data = ""

        if not isinstance(data, str):
            raise InvalidFileExtension("Data must be string type.")

        with open(Path(storage_path) / filename, "w+") as f:
            f.write(data)

    else:
        raise InvalidFileExtension("The file extension is not supported.")


def load_file(storage_path: str, filename: str) -> Any:
    """This function loads data for the filenames in the DB."""

    if filename.endswith(".pkl"):
        with open(Path(storage_path) / filename, "rb") as f:
            data = cloudpickle.load(f)

    elif filename.endswith(".log") or filename.endswith(".txt"):
        with open(Path(storage_path) / filename, "r") as f:
            data = f.read()

    return data



import os
import urllib.request
from pathlib import Path
from subprocess import PIPE, CalledProcessError, Popen
from typing import Optional

from furl import furl

from covalent._file_transfer import File
from covalent._file_transfer.strategies.transfer_strategy_base import FileTransferStrategy


class S3(FileTransferStrategy):

    """
    Implements Base FileTransferStrategy class to upload/download files from S3 Bucket.
    """

    def __init__(
        self,
        credentials: str = os.environ.get("AWS_SHARED_CREDENTIALS_FILE")
        or os.path.join(os.environ["HOME"], ".aws/credentials"),
        profile: str = os.environ.get("AWS_PROFILE") or None,
        region_name: str = os.getenv("AWS_REGION") or None,
    ):

        self.credentials = credentials
        self.profile = profile
        self.region_name = region_name

        try:
            import boto3
        except ImportError:
            raise ImportError(
                "Using S3 strategy requires boto3 from AWS installed on your system."
            )

        os.environ["AWS_SHARED_CREDENTIALS_FILE"] = self.credentials
        if self.profile is not None:
            os.environ["AWS_PROFILE"] = self.profile

        # AWS Account Retrieval
        sts = boto3.client("sts")
        identity = sts.get_caller_identity()
        account = identity.get("Account")

        if account is None:
            raise Exception("Incorrect AWS account credentials")

    # return callable to download here implies 'from' is a remote source
    def download(self, from_file: File, to_file: File = File()) -> File:
        from_filepath = from_file.filepath
        to_filepath = to_file.filepath
        bucket_name = furl(from_file.uri).origin[5:]

        def callable():
            import boto3

            s3 = boto3.client(
                "s3",
                region_name=self.region_name,
            )
            s3.download_file(bucket_name, from_filepath, to_filepath)

        return callable

    # return callable to download here implies 'to' is a remote source
    def upload(self, from_file: File, to_file: File = File()) -> File:

        from_filepath = from_file.filepath
        to_filepath = to_file.filepath
        bucket_name = furl(to_file.uri).origin[5:]

        def callable():
            import boto3

            s3 = boto3.client(
                "s3",
                region_name=self.region_name,
            )
            s3.upload_file(from_filepath, bucket_name, to_filepath)

        return callable

    # No S3 Strategy for copy
    def cp(self, from_file: File, to_file: File = File()) -> File:
        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
# Relief from the License may be granted by purchasing a commercial license.

"""Dispatch request and response model"""


from datetime import datetime
from enum import Enum
from typing import List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, conint

from covalent_ui.api.v1.utils.status import Status


class CaseInsensitiveEnum(Enum):
    """Enum overriden to support case insensitive keys"""

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value == value.upper():
                return member


class SortBy(CaseInsensitiveEnum):
    """Values to filter data by"""

    RUNTIME = "runtime"
    STATUS = "status"
    STARTED = "started_at"
    LATTICE_NAME = "lattice_name"
    ENDED = "ended_at"


class SortDirection(CaseInsensitiveEnum):
    """Values to decide sort direction"""

    ASCENDING = "ASC"
    DESCENDING = "DESC"


class DispatchSummaryRequest(BaseModel):
    count: conint(gt=0, lt=100)
    offset: Optional[conint(gt=-1)] = 0
    sort_by: Optional[SortBy] = SortBy.STARTED
    search: Optional[str] = ""
    direction: Optional[SortDirection] = SortDirection.DESCENDING


class DispatchModule(BaseModel):
    """Dispatch Modeule Validation"""

    dispatch_id: str
    lattice_name: str
    runtime: Optional[Union[int, None]]
    total_electrons: Optional[Union[int, None]]
    total_electrons_completed: Optional[Union[int, None]]
    started_at: Optional[Union[datetime, None]]
    ended_at: Optional[Union[datetime, None]]
    status: Status
    updated_at: Optional[Union[datetime, None]]


class DispatchResponse(BaseModel):
    """Dispatch Response Model"""

    items: List[DispatchModule]
    total_count: int

    class Config:
        """Configure example for openAPI"""

        schema_extra = {
            "example": {
                "dispatches": [
                    {
                        "dispatch_id": "1b44989a-1c65-4148-959e-00062a34ac16",
                        "lattice_name": "testing content",
                        "runtime": 1,
                        "started_time": "2022-06-13T07:45:02.114328+00:00",
                        "end_time": "2022-06-13T07:45:02.216474+00:00",
                        "status": "COMPLETED",
                    }
                ],
                "total_count": 10,
            }
        }


class DeleteDispatchesRequest(BaseModel):
    """Dashboard metadate model"""

    dispatches: List[UUID]


class DeleteDispatchesResponse(BaseModel):
    """Dashboard metadate model"""

    success_items: List[UUID]
    failure_items: List[UUID] = None
    message: str = None


class DispatchDashBoardResponse(BaseModel):
    """Dashboard metadate model"""

    total_jobs: int = None
    total_jobs_running: int = None
    total_jobs_completed: int = None
    total_jobs_failed: int = None
    latest_running_task_status: Status = None
    total_dispatcher_duration: int = None

    class Config:
        """Configure example for openAPI"""

        schema_extra = {
            "example": {
                "total_jobs_running": 5,
                "total_jobs_done": 20,
                "latest_running_task_status": "COMPLETED",
                "total_dispatcher_duration": 90,
            }
        }



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Graph Data Layer"""
from uuid import UUID

from fastapi import HTTPException
from sqlalchemy.orm import Session

from covalent_ui.api.v1.database.schema.electron import Electron
from covalent_ui.api.v1.database.schema.electron_dependency import ElectronDependency
from covalent_ui.api.v1.database.schema.lattices import Lattice
from covalent_ui.api.v1.models.lattices_model import GraphResponse


class Graph:
    """Graph data access layer"""

    def __init__(self, db_con: Session) -> None:
        self.db_con = db_con

    def get_nodes(self, dispatch_id: UUID):
        """
        Get nodes from dispatch id
        When dispatch id passed to get graph
            Join lattice and electron on Electron.parent_lattice_id == Lattice.id
            then filter it with dispatch id
        Args:
            dispatch_id: Refers to the dispatch id from lattices table
        Return:
            graph data with list of nodes and links
        """
        return (
            self.db_con.query(
                Electron.id,
                Electron.name,
                Electron.transport_graph_node_id.label("node_id"),
                Electron.started_at,
                Electron.completed_at,
                Electron.status,
                Electron.type,
            )
            .filter(
                Electron.parent_lattice_id == Lattice.id, Lattice.dispatch_id == str(dispatch_id)
            )
            .all()
        )

    def get_links(self, dispatch_id: UUID):
        """
        Get links from dispatch id
        When dispatch id passed to get graph
            Join lattice and electron by Electron.parent_lattice_id == Lattice.id
            Join Electron and Electron Dependency by ElectronDependency.electron_id == Electron.id
            then filter it with dispatch id
        Args:
            dispatch_id: Refers to the dispatch id from lattices table
        Return:
            graph data with list of nodes and links
        """
        return (
            self.db_con.query(
                ElectronDependency.edge_name,
                ElectronDependency.parameter_type,
                ElectronDependency.electron_id.label("target"),
                ElectronDependency.parent_electron_id.label("source"),
                ElectronDependency.arg_index,
            )
            .filter(
                Electron.parent_lattice_id == Lattice.id,
                ElectronDependency.electron_id == Electron.id,
                Lattice.dispatch_id == str(dispatch_id),
            )
            .all()
        )

    def check_error(self, data):
        """
        Helper method to rise exception if data is None

        Args:
            data: list of queried data
        Return:
            data
        Rise:
            Http Exception with status code 400 and details
        """
        if data is None:
            raise HTTPException(status_code=400, detail=["Something went wrong"])
        return data

    def get_graph(self, dispatch_id: UUID):
        """
        Get graph data from parent lattice id
        When dispatch id passed to get graph
            Get list of nodes from Electrons table by passing list of latice id with the dispatch id
            Get list of links from Electron dependency table by passing in electron ids
        Args:
            dispatch_id: Refers to the dispatch id from lattices table
        Return:
            graph data with list of nodes and links
        """
        data = self.get_nodes(dispatch_id=dispatch_id)
        if len(data) == 0:
            raise HTTPException(
                status_code=400,
                detail=[
                    {
                        "loc": ["path", "dispatch_id"],
                        "msg": f"Dispatch ID {dispatch_id} does not exist",
                        "type": None,
                    }
                ],
            )
        nodes = self.check_error(data=data)
        data = self.get_links(dispatch_id=dispatch_id)
        links = self.check_error(data=data)
        return GraphResponse(dispatch_id=str(dispatch_id), graph={"nodes": nodes, "links": links})



import os
import urllib.request
from pathlib import Path
from subprocess import PIPE, CalledProcessError, Popen
from typing import Optional

from covalent._file_transfer import File
from covalent._file_transfer.strategies.transfer_strategy_base import FileTransferStrategy


class HTTP(FileTransferStrategy):
    """
    Implements Base FileTransferStrategy class to use HTTP to download files from public URLs.
    """

    # return callable to download here implies 'from' is a remote source
    def download(self, from_file: File, to_file: File = File()) -> File:
        from_filepath = from_file.uri
        to_filepath = to_file.filepath

        def callable():
            urllib.request.urlretrieve(from_filepath, to_filepath)
            return to_filepath

        return callable

    # HTTP Strategy is read only
    def upload(self, from_file: File, to_file: File = File()) -> File:
        raise NotImplementedError

    # HTTP Strategy is read only
    def cp(self, from_file: File, to_file: File = File()) -> File:
        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""
Module for defining a local executor that directly invokes the input python function.

This is a plugin executor module; it is loaded if found and properly structured.
"""

import io
import os
import subprocess
from contextlib import redirect_stderr, redirect_stdout
from typing import Any, Callable, Dict, List

# Relative imports are not allowed in executor plugins
from covalent._shared_files import logger
from covalent._shared_files.util_classes import DispatchInfo
from covalent._workflow.transport import TransportableObject
from covalent.executor import BaseExecutor, wrapper_fn

# The plugin class name must be given by the executor_plugin_name attribute:
EXECUTOR_PLUGIN_NAME = "LocalExecutor"

app_log = logger.app_log
log_stack_info = logger.log_stack_info

_EXECUTOR_PLUGIN_DEFAULTS = {
    "log_stdout": "stdout.log",
    "log_stderr": "stderr.log",
    "cache_dir": os.path.join(
        os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"), "covalent"
    ),
}


class LocalExecutor(BaseExecutor):
    """
    Local executor class that directly invokes the input function.
    """

    def run(self, function: Callable, args: List, kwargs: Dict, task_metadata: Dict):
        app_log.debug(f"Running function {function} locally")
        return function(*args, **kwargs)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Lattice request and response model"""

from datetime import datetime
from enum import Enum

from pydantic import BaseModel


class LatticeDetailResponse(BaseModel):
    """Lattices details model"""

    dispatch_id: str = None
    status: str = None
    total_electrons: int = None
    total_electrons_completed: int = None
    started_at: datetime = None
    ended_at: datetime = None
    directory: str = None
    runtime: int = None
    updated_at: datetime = None


class LatticeFileResponse(BaseModel):
    """Lattices File Response Model"""

    data: str = None


class LatticeExecutorResponse(BaseModel):
    """Lattices File Response Model"""

    executor_name: str = None
    executor_details: dict = None


class LatticeWorkflowExecutorResponse(BaseModel):
    """Lattices File Response Model"""

    workflow_executor_name: str = None
    workflow_executor_details: dict = None


class GraphResponse(BaseModel):
    """Graph Response Model"""

    dispatch_id: str = None
    graph: dict = None


class ElectronResponse(BaseModel):
    """Electron Response Model"""

    id: int = None
    node_id: int = None
    parent_lattice_id: int = None
    type: str = None
    storage_path: str = None
    name: str = None
    status: str = None
    started_at: datetime = None
    ended_at: datetime = None
    runtime: int = None


class ElectronFileResponse(BaseModel):
    """Electron Response Model"""

    data: str = None


class ElectronExecutorResponse(BaseModel):
    """Lattices File Response Model"""

    executor_name: str = None
    executor_details: dict = None


class ElectronErrorResponse(BaseModel):
    """Eelctron Error Response Model"""

    data: str = None


class ElectronFunctionResponse(BaseModel):
    """Electron Function Response Model"""

    data: str = None


class FileOutput(str, Enum):
    RESULT = "result"
    FUNCTION_STRING = "function_string"
    INPUTS = "inputs"
    ERROR = "error"
    EXECUTOR = "executor"
    WORKFLOW_EXECUTOR = "workflow_executor"
    FUNCTION = "function"
    TRANSPORT_GRAPH = "transport_graph"



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Lattice file module mapper"""

from enum import Enum


class Filetype(str, Enum):
    """Map module for mapping file module to their
    corresponding file extensions"""

    RESULT = "result.pkl"
    OUTPUT = "output"
    FUNCTION_STRING = "function_string.txt"
    INPUTS = "inputs.pkl"
    ERROR = "error.log"
    EXECUTOR = "executor.pkl"
    RESULTS = "results.pkl"


class FileMapper(str, Enum):
    """Map module for mapping file module to their
    corresponding result from database"""

    ERROR = 3
    FUNCTION_STRING = 4
    EXECUTOR = 5
    INPUTS = 6
    RESULT = 7


class FileExtension(Enum):
    """Extension enum"""

    TXT = ".txt"
    PKL = ".pkl"



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Import workflow utilities."""

from .depsbash import DepsBash
from .depscall import DepsCall
from .depspip import DepsPip
from .electron import electron
from .lattice import lattice
from .lepton import Lepton
from .transport import TransportableObject



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import json

import aiohttp
import requests

import covalent_ui.app as ui_server
from covalent._results_manager import Result
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent_dispatcher._db.dispatchdb import encode_dict, extract_graph, extract_metadata

DEFAULT_PORT = get_config("user_interface.port")

app_log = logger.app_log

# UI server webhook for result updates


def get_ui_url(path):
    baseUrl = f"http://localhost:{DEFAULT_PORT}"
    return f"{baseUrl}{path}"


async def send_update(result: Result) -> None:
    """
    Signal UI server about a result update. Note that the server will expect the
    updated result to have been saved to the results directory prior to the
    update.

    Args: result: The updated result object.

    Returns: None
    """

    result_update = json.dumps(
        {
            "event": "result-update",
            "result": {
                "dispatch_id": result.dispatch_id,
                "results_dir": result.results_dir,
                "status": result.status.STATUS,
            },
        },
    )

    app_log.debug("Moving to Fast API soon - stay tuned!!")
    try:
        # ignore response
        timeout = aiohttp.ClientTimeout(total=1)
        async with aiohttp.ClientSession(timeout=timeout) as session:
            async with session.post(
                get_ui_url(ui_server.WEBHOOK_PATH), json=json.loads(result_update)
            ) as resp:
                status = resp.status
                text = await resp.text()
                app_log.debug(f"send_update received response {status}, {text}")
    except Exception as ex:
        # catch all requests-related exceptions
        app_log.debug(f"Unable to send result update to UI server: {ex}")


def send_draw_request(lattice) -> None:
    """
    Sends a lattice draw request to UI server along with all necessary lattice
    graph data.

    Args: lattice: The lattice to draw with a pre-built graph.

    Returns: None
    """

    graph = lattice.transport_graph.get_internal_graph_copy()

    named_args = {k: v.object_string for k, v in lattice.named_args.items()}
    named_kwargs = {k: v.object_string for k, v in lattice.named_kwargs.items()}

    draw_request = json.dumps(
        {
            "event": "draw-request",
            "payload": {
                "lattice": {
                    "function_string": lattice.workflow_function_string,
                    "doc": lattice.__doc__,
                    "name": lattice.__name__,
                    "inputs": encode_dict({**named_args, **named_kwargs}),
                    "metadata": extract_metadata(lattice.metadata),
                },
                "graph": extract_graph(graph),
            },
        }
    )

    try:
        response = requests.post(get_ui_url("/api/draw"), data=draw_request)
        response.raise_for_status()
    except requests.exceptions.HTTPError as ex:
        app_log.error(ex)
    except requests.exceptions.RequestException:
        app_log.error("Connection failure. Please check Covalent server is running.")



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import codecs
import json
from typing import Optional
from uuid import UUID

import cloudpickle as pickle
from fastapi import APIRouter, Request
from fastapi.responses import JSONResponse
from sqlalchemy.orm import Session

import covalent_dispatcher as dispatcher
from covalent._data_store.datastore import workflow_db
from covalent._data_store.models import Lattice
from covalent._results_manager.result import Result
from covalent._results_manager.results_manager import result_from
from covalent._shared_files import logger

from .._db.dispatchdb import DispatchDB

app_log = logger.app_log
log_stack_info = logger.log_stack_info

router: APIRouter = APIRouter()


@router.post("/submit")
async def submit(request: Request) -> UUID:
    """
    Function to accept the submit request of
    new dispatch and return the dispatch id
    back to the client.

    Args:
        None

    Returns:
        dispatch_id: The dispatch id in a json format
                     returned as a Fast API Response object.
    """
    data = await request.json()
    data = json.dumps(data).encode("utf-8")

    dispatch_id = await dispatcher.run_dispatcher(data)
    return dispatch_id


@router.post("/cancel")
async def cancel(request: Request) -> str:
    """
    Function to accept the cancel request of
    a dispatch.

    Args:
        None

    Returns:
        Fast API Response object confirming that the dispatch
        has been cancelled.
    """
    data = await request.body()
    dispatch_id = data.decode("utf-8")

    dispatcher.cancel_running_dispatch(dispatch_id)
    return f"Dispatch {dispatch_id} cancelled."


@router.get("/db-path")
def db_path() -> str:
    db_path = DispatchDB()._dbpath
    return json.dumps(db_path)


@router.get("/result/{dispatch_id}")
def get_result(
    dispatch_id: str, wait: Optional[bool] = False, status_only: Optional[bool] = False
):
    with workflow_db.session() as session:
        lattice_record = session.query(Lattice).where(Lattice.dispatch_id == dispatch_id).first()
        status = lattice_record.status if lattice_record else None
        if not lattice_record:
            return JSONResponse(
                status_code=404,
                content={"message": f"The requested dispatch ID {dispatch_id} was not found."},
            )
        if not wait or status in [
            str(Result.COMPLETED),
            str(Result.FAILED),
            str(Result.CANCELLED),
            str(Result.POSTPROCESSING_FAILED),
            str(Result.PENDING_POSTPROCESSING),
        ]:
            output = {
                "id": dispatch_id,
                "status": lattice_record.status,
            }
            if not status_only:
                output["result"] = codecs.encode(
                    pickle.dumps(result_from(lattice_record)), "base64"
                ).decode()
            return output

        response = JSONResponse(
            status_code=503,
            content={
                "message": "Result not ready to read yet. Please wait for a couple of seconds."
            },
            headers={"Retry-After": "2"},
        )
        return response



from typing import Optional

from covalent._file_transfer.file import File


class Folder(File):
    """
    Folder class to store components of provided URI including scheme (s3://, file://, ect.), determine if the file is remote,
    and act as facade to facilitate filesystem operations. Folder is a child of the File class which sets `is_dir` flag to True.

    Attributes:
        include_folder: Flag that determines if the folder should be included in the file transfer, if False only contents of folder are transfered.
    """

    def __init__(
        self,
        filepath: Optional[str] = None,
        is_remote: bool = False,
        is_dir: bool = True,
        include_folder: bool = False,
    ):
        super().__init__(filepath, is_remote, is_dir, include_folder)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""
Module for defining a Dask executor that submits the input python function in a dask cluster
and waits for execution to finish then returns the result.

This is a plugin executor module; it is loaded if found and properly structured.
"""

import io
import os
import sys
from contextlib import redirect_stderr, redirect_stdout
from typing import Callable, Dict, List

from dask.distributed import Client

from covalent._shared_files import logger

# Relative imports are not allowed in executor plugins
from covalent._shared_files.config import get_config
from covalent._shared_files.utils import _address_client_mapper
from covalent.executor.base import BaseAsyncExecutor

# The plugin class name must be given by the executor_plugin_name attribute:
EXECUTOR_PLUGIN_NAME = "DaskExecutor"

app_log = logger.app_log
log_stack_info = logger.log_stack_info

_EXECUTOR_PLUGIN_DEFAULTS = {
    "log_stdout": "stdout.log",
    "log_stderr": "stderr.log",
    "cache_dir": os.path.join(
        os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"), "covalent"
    ),
}


def dask_wrapper(fn, args, kwargs):
    with redirect_stdout(io.StringIO()) as stdout, redirect_stderr(io.StringIO()) as stderr:
        output = fn(*args, **kwargs)

    return output, stdout.getvalue(), stderr.getvalue()


class DaskExecutor(BaseAsyncExecutor):
    """
    Dask executor class that submits the input function to a running dask cluster.
    """

    def __init__(
        self,
        scheduler_address: str = "",
        log_stdout: str = "stdout.log",
        log_stderr: str = "stderr.log",
        conda_env: str = "",
        cache_dir: str = "",
        current_env_on_conda_fail: bool = False,
    ) -> None:
        if not cache_dir:
            cache_dir = os.path.join(
                os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"),
                "covalent",
            )

        if not scheduler_address:
            try:
                scheduler_address = get_config("dask.scheduler_address")
            except KeyError as ex:
                app_log.debug(
                    "No dask scheduler address found in config. Address must be set manually."
                )

        super().__init__(log_stdout, log_stderr, conda_env, cache_dir, current_env_on_conda_fail)

        self.scheduler_address = scheduler_address

    async def run(self, function: Callable, args: List, kwargs: Dict, task_metadata: Dict):
        """Submit the function and inputs to the dask cluster"""

        node_id = task_metadata["node_id"]

        dask_client = _address_client_mapper.get(self.scheduler_address)

        if not dask_client:
            dask_client = Client(address=self.scheduler_address, asynchronous=True)
            _address_client_mapper[self.scheduler_address] = dask_client

            await dask_client

        future = dask_client.submit(dask_wrapper, function, args, kwargs)
        app_log.debug(f"Submitted task {node_id} to dask")
        result, worker_stdout, worker_stderr = await dask_client.gather(future)

        print(worker_stdout, end="", file=sys.stdout)
        print(worker_stderr, end="", file=sys.stderr)

        # FIX: need to get stdout and stderr from dask worker and print them
        return result



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Request and response models for Electrons"""

from enum import Enum

from pydantic import BaseModel

from covalent_ui.api.v1.utils.status import Status


class ElectronResponseModel(BaseModel):
    """Electron Response Validation"""

    id: int
    parent_lattice_id: int
    transport_graph_node_id: int
    type: str
    name: str
    status: Status
    executor: str
    created_at: str
    started_at: str
    completed_at: str
    updated_at: str


class FileOutput(str, Enum):
    FUNCTION_STRING = "function_string"
    FUNCTION = "function"
    EXECUTOR = "executor"
    RESULT = "result"
    VALUE = "value"
    KEY = "key"
    STDOUT = "stdout"
    DEPS = "deps"
    CALL_BEFORE = "call_before"
    CALL_AFTER = "call_after"
    ERROR = "error"
    INFO = "info"
    INPUTS = "inputs"



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Result object."""

import os
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Set, Union

from sqlalchemy import and_, update

from .._data_store import DataStore, DataStoreNotInitializedError, models, workflow_db
from .._shared_files import logger
from .._shared_files.context_managers import active_lattice_manager
from .._shared_files.defaults import prefix_separator, sublattice_prefix
from .._shared_files.util_classes import RESULT_STATUS, Status
from .._workflow.transport import TransportableObject
from .write_result_to_db import (
    get_electron_type,
    insert_electron_dependency_data,
    insert_electrons_data,
    insert_lattices_data,
    load_file,
    store_file,
    update_electrons_data,
    update_lattice_completed_electron_num,
    update_lattices_data,
)

if TYPE_CHECKING:
    from .._shared_files.util_classes import Status
    from .._workflow.lattice import Lattice

app_log = logger.app_log
log_stack_info = logger.log_stack_info

LATTICE_FUNCTION_FILENAME = "function.pkl"
LATTICE_FUNCTION_STRING_FILENAME = "function_string.txt"
LATTICE_EXECUTOR_DATA_FILENAME = "executor_data.pkl"
LATTICE_WORKFLOW_EXECUTOR_DATA_FILENAME = "workflow_executor_data.pkl"
LATTICE_ERROR_FILENAME = "error.log"
LATTICE_INPUTS_FILENAME = "inputs.pkl"
LATTICE_NAMED_ARGS_FILENAME = "named_args.pkl"
LATTICE_NAMED_KWARGS_FILENAME = "named_kwargs.pkl"
LATTICE_RESULTS_FILENAME = "results.pkl"
LATTICE_TRANSPORT_GRAPH_FILENAME = "transport_graph.pkl"
LATTICE_STORAGE_TYPE = "local"

ELECTRON_FUNCTION_FILENAME = "function.pkl"
ELECTRON_FUNCTION_STRING_FILENAME = "function_string.txt"
ELECTRON_VALUE_FILENAME = "value.pkl"
ELECTRON_EXECUTOR_DATA_FILENAME = "executor_data.pkl"
ELECTRON_STDOUT_FILENAME = "stdout.log"
ELECTRON_STDERR_FILENAME = "stderr.log"
ELECTRON_INFO_FILENAME = "info.log"
ELECTRON_RESULTS_FILENAME = "results.pkl"
ELECTRON_DEPS_FILENAME = "deps.pkl"
ELECTRON_CALL_BEFORE_FILENAME = "call_before.pkl"
ELECTRON_CALL_AFTER_FILENAME = "call_after.pkl"
ELECTRON_STORAGE_TYPE = "local"


class Result:
    """
    Result class to store and perform operations on the result obtained from a dispatch.

    Attributes:
        lattice: "Lattice" object which was dispatched.
        results_dir: Directory where the result will be stored.
                     It'll be in the format of "<results_dir>/<dispatch_id>/".
        dispatch_id: Dispatch id assigned to this dispatch.
        status: Status of the result. It'll be one of the following:
                - Result.NEW_OBJ: When it is a new result object.
                - Result.COMPLETED: When processing of all the nodes has completed successfully.
                - Result.RUNNING: When some node executions are in process.
                - Result.FAILED: When one or more node executions have failed.
                - Result.CANCELLED: When the dispatch was cancelled.
        result: Final result of the dispatch, i.e whatever the "Lattice" was returning as a function.
        inputs: Inputs sent to the "Lattice" function for dispatching.
        error: Error due to which the execution failed.

    Functions:
        save_result: Save the result object to the passed results directory or to self.results_dir by default.
        get_all_node_outputs: Return all the outputs of all the node executions.
    """

    NEW_OBJ = RESULT_STATUS.NEW_OBJECT
    COMPLETED = RESULT_STATUS.COMPLETED
    POSTPROCESSING = RESULT_STATUS.POSTPROCESSING
    PENDING_POSTPROCESSING = RESULT_STATUS.PENDING_POSTPROCESSING
    POSTPROCESSING_FAILED = RESULT_STATUS.POSTPROCESSING_FAILED
    RUNNING = RESULT_STATUS.RUNNING
    FAILED = RESULT_STATUS.FAILED
    CANCELLED = RESULT_STATUS.CANCELLED

    def __init__(self, lattice: "Lattice", results_dir: str, dispatch_id: str = None) -> None:

        self._start_time = None
        self._end_time = None

        self._results_dir = results_dir

        self._lattice = lattice
        self._dispatch_id = dispatch_id

        self._status = Result.NEW_OBJ

        self._result = TransportableObject(None)

        self._num_nodes = -1

        self._inputs = {"args": [], "kwargs": {}}
        if lattice.args:
            self._inputs["args"] = lattice.args
        if lattice.kwargs:
            self._inputs["kwargs"] = lattice.kwargs

        self._error = None

    def __str__(self):
        """String representation of the result object"""

        arg_str_repr = [e.object_string for e in self.inputs["args"]]
        kwarg_str_repr = {key: value.object_string for key, value in self.inputs["kwargs"].items()}

        show_result_str = f"""
Lattice Result
==============
status: {self._status}
result: {self.result}
input args: {arg_str_repr}
input kwargs: {kwarg_str_repr}
error: {self.error}

start_time: {self.start_time}
end_time: {self.end_time}

results_dir: {self.results_dir}
dispatch_id: {self.dispatch_id}

Node Outputs
------------
"""

        node_outputs = self.get_all_node_outputs()
        for k, v in node_outputs.items():
            show_result_str += f"{k}: {v.object_string}\n"

        return show_result_str

    @property
    def start_time(self) -> datetime:
        """
        Start time of processing the dispatch.
        """

        return self._start_time

    @property
    def end_time(self) -> datetime:
        """
        End time of processing the dispatch.
        """

        return self._end_time

    @property
    def results_dir(self) -> str:
        """
        Results directory used to save this result object.
        """

        return self._results_dir

    @property
    def lattice(self) -> "Lattice":
        """
        "Lattice" object which was dispatched.
        """

        return self._lattice

    @property
    def dispatch_id(self) -> str:
        """
        Dispatch id of current dispatch.
        """

        return self._dispatch_id

    @property
    def status(self) -> Status:
        """
        Status of current dispatch.
        """

        return self._status

    @property
    def encoded_result(self) -> TransportableObject:
        """
        Encoded final result of current dispatch
        """
        return self._result

    @property
    def result(self) -> Union[int, float, list, dict]:
        """
        Final result of current dispatch.
        """

        return self._result.get_deserialized()

    @property
    def inputs(self) -> dict:
        """
        Inputs sent to the "Lattice" function for dispatching.
        """

        return self._inputs

    @property
    def error(self) -> str:
        """
        Error due to which the dispatch failed.
        """

        return self._error

    def _initialize_nodes(self) -> None:
        """
        Initialize the nodes of the transport graph with a blank result.
        This is called after `self.lattice.transport_graph` has been deserialized.

        Args:
            None

        Returns:
            None
        """

        self._num_nodes = self.lattice.transport_graph.get_internal_graph_copy().number_of_nodes()
        for node_id in range(self._num_nodes):
            node_name = self.lattice.transport_graph.get_node_value(node_id, "name")

            self.lattice.transport_graph.set_node_value(node_id, "node_name", node_name)

            self.lattice.transport_graph.set_node_value(node_id, "start_time", None)

            self.lattice.transport_graph.set_node_value(node_id, "end_time", None)

            self.lattice.transport_graph.set_node_value(node_id, "status", Result.NEW_OBJ)

            self.lattice.transport_graph.set_node_value(node_id, "output", None)

            self.lattice.transport_graph.set_node_value(node_id, "error", None)

            self.lattice.transport_graph.set_node_value(node_id, "sublattice_result", None)

            self.lattice.transport_graph.set_node_value(node_id, "stdout", None)

            self.lattice.transport_graph.set_node_value(node_id, "stderr", None)

    def get_node_result(self, node_id: int) -> dict:
        """Return the result of a particular node.

        Args:
            node_id: The node id.

        Returns:
            node_result: The result of the node containing below in a dictionary format:
                            - node_id: The node id.
                            - node_name: The name of the node.
                            - start_time: The start time of the node execution.
                            - end_time: The end time of the node execution.
                            - status: The status of the node execution.
                            - output: The output of the node unless error occured in which case None.
                            - error: The error of the node if occured else None.
                            - sublattice_result: The result of the sublattice if any.
                            - stdout: The stdout of the node execution.
                            - stderr: The stderr of the node execution.
        """

        return {
            "node_id": node_id,
            "node_name": self._get_node_name(node_id),
            "start_time": self.lattice.transport_graph.get_node_value(node_id, "start_time"),
            "end_time": self.lattice.transport_graph.get_node_value(node_id, "end_time"),
            "status": self._get_node_status(node_id),
            "output": self._get_node_output(node_id),
            "error": self.lattice.transport_graph.get_node_value(node_id, "error"),
            "sublattice_result": self.lattice.transport_graph.get_node_value(
                node_id, "sublattice_result"
            ),
            "stdout": self.lattice.transport_graph.get_node_value(node_id, "stdout"),
            "stderr": self._get_node_error(node_id),
        }

    def get_all_node_outputs(self) -> dict:
        """
        Return output of every node execution.

        Args:
            None

        Returns:
            node_outputs: A dictionary containing the output of every node execution.
        """
        with workflow_db.session() as session:

            lattice_id = (
                session.query(models.Lattice)
                .where(models.Lattice.dispatch_id == self.dispatch_id)
                .first()
                .id
            )
            electron_records = (
                session.query(models.Electron)
                .where(models.Electron.parent_lattice_id == lattice_id)
                .all()
            )
            all_node_outputs = {}
            for electron in electron_records:
                node_id = electron.transport_graph_node_id
                all_node_outputs[
                    f"{self._get_node_name(node_id=node_id)}({node_id})"
                ] = self._get_node_output(node_id=node_id)
            return all_node_outputs

    def get_all_node_results(self) -> List[Dict]:
        """
        Get all the node results.

        Args:
            None

        Returns:
            node_results: A list of dictionaries containing the result of every node execution.
        """
        with workflow_db.session() as session:

            lattice_id = (
                session.query(models.Lattice)
                .where(models.Lattice.dispatch_id == self.dispatch_id)
                .first()
                .id
            )
            electron_records = (
                session.query(models.Electron)
                .where(models.Electron.parent_lattice_id == lattice_id)
                .all()
            )
            return [
                self.get_node_result(node_id=electron.transport_graph_node_id)
                for electron in electron_records
            ]

    def post_process(self):

        # Copied from server-side _post_process()
        node_outputs = self.get_all_node_outputs()
        ordered_node_outputs = []
        for i, item in enumerate(node_outputs.items()):
            key, val = item
            if not key.startswith(prefix_separator) or key.startswith(sublattice_prefix):
                ordered_node_outputs.append((i, val))

        lattice = self._lattice

        with active_lattice_manager.claim(lattice):
            lattice.post_processing = True
            lattice.electron_outputs = ordered_node_outputs
            workflow_function = lattice.workflow_function.get_deserialized()
            result = workflow_function(*lattice.args, **lattice.kwargs)
            lattice.post_processing = False
        return result

    def _get_node_name(self, node_id: int) -> str:
        """
        Returns the name of the node with given node id.

        Args:
            node_id: The node id.

        Returns:
            The name of said node.
        """

        with workflow_db.session() as session:

            lattice_id = (
                session.query(models.Lattice)
                .where(models.Lattice.dispatch_id == self.dispatch_id)
                .first()
                .id
            )
            return (
                session.query(models.Electron)
                .where(
                    and_(
                        models.Electron.parent_lattice_id == lattice_id,
                        models.Electron.transport_graph_node_id == node_id,
                    )
                )
                .first()
                .name
            )

    def _get_node_status(self, node_id: int) -> "Status":
        """
        Returns the status of a node.

        Args:
            node_id: The node id.

        Returns:
            The status of said node.
        """

        with workflow_db.session() as session:

            lattice_id = (
                session.query(models.Lattice)
                .where(models.Lattice.dispatch_id == self.dispatch_id)
                .first()
            ).id
            return (
                session.query(models.Electron)
                .where(
                    and_(
                        models.Electron.parent_lattice_id == lattice_id,
                        models.Electron.transport_graph_node_id == node_id,
                    )
                )
                .first()
                .status
            )

    def _get_node_output(self, node_id: int) -> Any:
        """
        Return the output of a node.

        Args:
            node_id: The node id.

        Returns:
            output: The output of said node.
                    Will return None if error occured in execution.
        """

        with workflow_db.session() as session:

            lattice_id = (
                session.query(models.Lattice)
                .where(models.Lattice.dispatch_id == self.dispatch_id)
                .first()
                .id
            )
            electron = (
                session.query(models.Electron)
                .where(
                    (
                        and_(
                            models.Electron.parent_lattice_id == lattice_id,
                            models.Electron.transport_graph_node_id == node_id,
                        )
                    )
                )
                .first()
            )
            return load_file(
                storage_path=electron.storage_path, filename=electron.results_filename
            )

    def _get_node_value(self, node_id: int) -> Any:
        """
        Return the output of a node.

        Args:
            node_id: The node id.

        Returns:
            output: The output of said node.
                    Will return None if error occured in execution.
        """

        with workflow_db.session() as session:

            lattice_id = (
                session.query(models.Lattice)
                .where(models.Lattice.dispatch_id == self.dispatch_id)
                .first()
                .id
            )
            electron = (
                session.query(models.Electron)
                .where(
                    (
                        and_(
                            models.Electron.parent_lattice_id == lattice_id,
                            models.Electron.transport_graph_node_id == node_id,
                        )
                    )
                )
                .first()
            )
            return load_file(storage_path=electron.storage_path, filename=electron.value_filename)

    def _get_node_error(self, node_id: int) -> Any:
        """
        Return the error of a node.

        Args:
            node_id: The node id.

        Returns:
            error: The error of said node.
                   Will return None if no error occured in execution.
        """

        with workflow_db.session() as session:

            lattice_id = (
                session.query(models.Lattice)
                .where(models.Lattice.dispatch_id == self.dispatch_id)
                .first()
                .id
            )
            electron = (
                session.query(models.Electron)
                .where(
                    (
                        and_(
                            models.Electron.parent_lattice_id == lattice_id,
                            models.Electron.transport_graph_node_id == node_id,
                        )
                    )
                )
                .first()
            )
            return load_file(storage_path=electron.storage_path, filename=electron.stderr_filename)

    def _update_node(
        self,
        node_id: int,
        node_name: str = None,
        start_time: "datetime" = None,
        end_time: "datetime" = None,
        status: "Status" = None,
        output: Any = None,
        error: Exception = None,
        sublattice_result: "Result" = None,
        stdout: str = None,
        stderr: str = None,
    ) -> None:
        """
        Update the node result in the transport graph.
        Called after any change in node's execution state.

        Args:
            node_id: The node id.
            node_name: The name of the node.
            start_time: The start time of the node execution.
            end_time: The end time of the node execution.
            status: The status of the node execution.
            output: The output of the node unless error occured in which case None.
            error: The error of the node if occured else None.
            sublattice_result: The result of the sublattice if any.
            stdout: The stdout of the node execution.
            stderr: The stderr of the node execution.

        Returns:
            None
        """

        app_log.debug("Inside update node")

        electron_kwargs = {}
        node_path = Path(self.results_dir) / self.dispatch_id / f"node_{node_id}"
        if not node_path.exists():
            node_path.mkdir()

        if node_name is not None:
            self.lattice.transport_graph.set_node_value(node_id, "node_name", node_name)
            electron_kwargs["name"] = node_name

        if start_time is not None:
            self.lattice.transport_graph.set_node_value(node_id, "start_time", start_time)
            electron_kwargs["started_at"] = start_time

        if end_time is not None:
            self.lattice.transport_graph.set_node_value(node_id, "end_time", end_time)
            electron_kwargs["completed_at"] = end_time

        if status is not None:
            self.lattice.transport_graph.set_node_value(node_id, "status", status)
            electron_kwargs["status"] = str(status)

        if output is not None:
            self.lattice.transport_graph.set_node_value(node_id, "output", output)
            store_file(node_path, ELECTRON_RESULTS_FILENAME, output)

        if error is not None:
            self.lattice.transport_graph.set_node_value(node_id, "error", error)

        if sublattice_result is not None:
            self.lattice.transport_graph.set_node_value(
                node_id, "sublattice_result", sublattice_result
            )

        if stdout is not None:
            self.lattice.transport_graph.set_node_value(node_id, "stdout", stdout)
            store_file(node_path, ELECTRON_STDOUT_FILENAME, stdout)

        if stderr is not None:
            self.lattice.transport_graph.set_node_value(node_id, "stderr", stderr)
            store_file(node_path, ELECTRON_STDERR_FILENAME, stderr)

        if str(status) == "COMPLETED":
            update_lattice_completed_electron_num(self.dispatch_id)

        with workflow_db.session() as session:
            lattice_id = (
                session.query(models.Lattice)
                .where(models.Lattice.dispatch_id == self.dispatch_id)
                .all()[0]
                .id
            )
            session.execute(
                update(models.Electron)
                .where(
                    models.Electron.parent_lattice_id == lattice_id,
                    models.Electron.transport_graph_node_id == node_id,
                )
                .values(updated_at=datetime.now(timezone.utc), **electron_kwargs)
            )
            session.commit()
        app_log.debug("Inside update node - SUCCESS")

    def _initialize_results_dir(self):
        """Create the results directory."""

        result_folder_path = os.path.join(self.results_dir, f"{self.dispatch_id}")
        Path(result_folder_path).mkdir(parents=True, exist_ok=True)

    def upsert_lattice_data(self):
        """Update lattice data"""

        with workflow_db.session() as session:
            lattice_exists = (
                session.query(models.Lattice)
                .where(models.Lattice.dispatch_id == self.dispatch_id)
                .first()
                is not None
            )

        try:
            workflow_func_string = self.lattice.workflow_function_string
        except AttributeError:
            workflow_func_string = None

        # Store all lattice info that belongs in filenames in the results directory
        data_storage_path = Path(self.results_dir) / self.dispatch_id
        for filename, data in [
            (LATTICE_FUNCTION_FILENAME, self.lattice.workflow_function),
            (LATTICE_FUNCTION_STRING_FILENAME, workflow_func_string),
            (LATTICE_EXECUTOR_DATA_FILENAME, self.lattice.metadata["executor_data"]),
            (
                LATTICE_WORKFLOW_EXECUTOR_DATA_FILENAME,
                self.lattice.metadata["workflow_executor_data"],
            ),
            (LATTICE_ERROR_FILENAME, self.error),
            (LATTICE_INPUTS_FILENAME, self.inputs),
            (LATTICE_NAMED_ARGS_FILENAME, self.lattice.named_args),
            (LATTICE_NAMED_KWARGS_FILENAME, self.lattice.named_kwargs),
            (LATTICE_RESULTS_FILENAME, self._result),
            (LATTICE_TRANSPORT_GRAPH_FILENAME, self._lattice.transport_graph),
        ]:

            store_file(data_storage_path, filename, data)

        # Write lattice records to Database
        if not lattice_exists:
            lattice_record_kwarg = {
                "dispatch_id": self.dispatch_id,
                "status": str(self.status),
                "name": self.lattice.__name__,
                "electron_num": self._num_nodes,
                "completed_electron_num": 0,  # None of the nodes have been executed or completed yet.
                "storage_path": str(data_storage_path),
                "storage_type": LATTICE_STORAGE_TYPE,
                "function_filename": LATTICE_FUNCTION_FILENAME,
                "function_string_filename": LATTICE_FUNCTION_STRING_FILENAME,
                "executor": self.lattice.metadata["executor"],
                "executor_data_filename": LATTICE_EXECUTOR_DATA_FILENAME,
                "workflow_executor": self.lattice.metadata["workflow_executor"],
                "workflow_executor_data_filename": LATTICE_WORKFLOW_EXECUTOR_DATA_FILENAME,
                "error_filename": LATTICE_ERROR_FILENAME,
                "inputs_filename": LATTICE_INPUTS_FILENAME,
                "named_args_filename": LATTICE_NAMED_ARGS_FILENAME,
                "named_kwargs_filename": LATTICE_NAMED_KWARGS_FILENAME,
                "results_filename": LATTICE_RESULTS_FILENAME,
                "transport_graph_filename": LATTICE_TRANSPORT_GRAPH_FILENAME,
                "created_at": datetime.now(timezone.utc),
                "updated_at": datetime.now(timezone.utc),
                "started_at": self.start_time,
                "completed_at": self.end_time,
            }
            insert_lattices_data(**lattice_record_kwarg)

        else:
            lattice_record_kwarg = {
                "dispatch_id": self.dispatch_id,
                "status": str(self.status),
                "electron_num": self._num_nodes,
                "updated_at": datetime.now(timezone.utc),
                "started_at": self.start_time,
                "completed_at": self.end_time,
            }
            update_lattices_data(**lattice_record_kwarg)

    def upsert_electron_data(self):
        """Update electron data"""

        tg = self.lattice.transport_graph
        dirty_nodes = set(tg.dirty_nodes)
        tg.dirty_nodes.clear()  # Ensure that dirty nodes list is reset once the data is updated
        app_log.debug("upsert_electron_data session begin")
        with workflow_db.session() as session:
            app_log.debug("upsert_electron_data session success")
            for node_id in dirty_nodes:

                node_path = Path(self.results_dir) / self.dispatch_id / f"node_{node_id}"

                if not node_path.exists():
                    node_path.mkdir()

                attribute_name = tg.get_node_value(node_key=node_id, value_key="name")

                try:
                    function_string = tg.get_node_value(node_id, "function_string")
                except KeyError:
                    function_string = None
                try:
                    node_value = tg.get_node_value(node_id, "value")
                except KeyError:
                    node_value = None
                try:
                    node_stdout = tg.get_node_value(node_id, "stdout")
                except KeyError:
                    node_stdout = None
                try:
                    node_stderr = tg.get_node_value(node_id, "stderr")
                except KeyError:
                    node_stderr = None
                try:
                    node_info = tg.get_node_value(node_id, "info")
                except KeyError:
                    node_info = None
                try:
                    node_output = tg.get_node_value(node_id, "output")
                except KeyError:
                    node_output = TransportableObject(None)
                if not isinstance(node_output, TransportableObject):
                    node_output = TransportableObject(node_output)

                executor = tg.get_node_value(node_id, "metadata")["executor"]
                started_at = tg.get_node_value(node_key=node_id, value_key="start_time")
                completed_at = tg.get_node_value(node_key=node_id, value_key="end_time")

                for filename, data in [
                    (ELECTRON_FUNCTION_FILENAME, tg.get_node_value(node_id, "function")),
                    (ELECTRON_FUNCTION_STRING_FILENAME, function_string),
                    (ELECTRON_VALUE_FILENAME, node_value),
                    (
                        ELECTRON_EXECUTOR_DATA_FILENAME,
                        tg.get_node_value(node_id, "metadata")["executor_data"],
                    ),
                    (ELECTRON_DEPS_FILENAME, tg.get_node_value(node_id, "metadata")["deps"]),
                    (
                        ELECTRON_CALL_BEFORE_FILENAME,
                        tg.get_node_value(node_id, "metadata")["call_before"],
                    ),
                    (
                        ELECTRON_CALL_AFTER_FILENAME,
                        tg.get_node_value(node_id, "metadata")["call_after"],
                    ),
                    (ELECTRON_STDOUT_FILENAME, node_stdout),
                    (ELECTRON_STDERR_FILENAME, node_stderr),
                    (ELECTRON_INFO_FILENAME, node_info),
                    (ELECTRON_RESULTS_FILENAME, node_output),
                ]:
                    store_file(node_path, filename, data)

                electron_exists = (
                    session.query(models.Electron, models.Lattice)
                    .where(
                        models.Electron.parent_lattice_id == models.Lattice.id,
                        models.Lattice.dispatch_id == self.dispatch_id,
                        models.Electron.transport_graph_node_id == node_id,
                    )
                    .first()
                    is not None
                )

                if not electron_exists:
                    electron_record_kwarg = {
                        "parent_dispatch_id": self.dispatch_id,
                        "transport_graph_node_id": node_id,
                        "type": get_electron_type(
                            tg.get_node_value(node_key=node_id, value_key="name")
                        ),
                        "name": tg.get_node_value(node_key=node_id, value_key="name"),
                        "status": str(tg.get_node_value(node_key=node_id, value_key="status")),
                        "storage_type": ELECTRON_STORAGE_TYPE,
                        "storage_path": str(node_path),
                        "function_filename": ELECTRON_FUNCTION_FILENAME,
                        "function_string_filename": ELECTRON_FUNCTION_STRING_FILENAME,
                        "executor": executor,
                        "executor_data_filename": ELECTRON_EXECUTOR_DATA_FILENAME,
                        "results_filename": ELECTRON_RESULTS_FILENAME,
                        "value_filename": ELECTRON_VALUE_FILENAME,
                        "stdout_filename": ELECTRON_STDOUT_FILENAME,
                        "stderr_filename": ELECTRON_STDERR_FILENAME,
                        "info_filename": ELECTRON_INFO_FILENAME,
                        "deps_filename": ELECTRON_DEPS_FILENAME,
                        "call_before_filename": ELECTRON_CALL_BEFORE_FILENAME,
                        "call_after_filename": ELECTRON_CALL_AFTER_FILENAME,
                        "created_at": datetime.now(timezone.utc),
                        "updated_at": datetime.now(timezone.utc),
                        "started_at": started_at,
                        "completed_at": completed_at,
                    }
                    insert_electrons_data(**electron_record_kwarg)

                else:
                    electron_record_kwarg = {
                        "parent_dispatch_id": self.dispatch_id,
                        "transport_graph_node_id": node_id,
                        "status": str(tg.get_node_value(node_key=node_id, value_key="status")),
                        "started_at": started_at,
                        "updated_at": datetime.now(timezone.utc),
                        "completed_at": completed_at,
                    }
                    update_electrons_data(**electron_record_kwarg)

    def insert_electron_dependency_data(self):
        """Update electron dependency data"""

        # Insert electron dependency records if they don't exist
        with workflow_db.session() as session:
            electron_dependencies_exist = (
                session.query(models.ElectronDependency, models.Electron, models.Lattice)
                .where(
                    models.Electron.id == models.ElectronDependency.electron_id,
                    models.Electron.parent_lattice_id == models.Lattice.id,
                    models.Lattice.dispatch_id == self.dispatch_id,
                )
                .first()
                is not None
            )
        app_log.debug("electron_dependencies_exist is " + str(electron_dependencies_exist))
        if not electron_dependencies_exist:
            insert_electron_dependency_data(dispatch_id=self.dispatch_id, lattice=self.lattice)

    def persist(self) -> None:
        """Save Result object to a DataStoreSession. Changes are queued until
        committed by the caller."""

        self._initialize_results_dir()
        app_log.debug("upsert start")
        self.upsert_lattice_data()
        self.upsert_electron_data()
        app_log.debug("upsert complete")
        self.insert_electron_dependency_data()
        app_log.debug("persis complete")

    def _convert_to_electron_result(self) -> Any:
        """
        Convert the result object to an electron's result.

        Args:
            None

        Returns:
            result: The final output of the dispatch.
        """

        return self._result


def _filter_cova_decorators(function_string: str, cova_imports: Set[str]) -> str:
    """
    Given a string representing a function, comment out any Covalent-related decorators.

    Args
        function_string: A string representation of a workflow function.

    Returns:
        The function string with Covalent-related decorators commented out.
    """

    has_cova_decorator = False
    in_decorator = 0
    function_lines = function_string.split("\n")
    for i in range(len(function_lines)):
        line = function_lines[i].strip()
        if in_decorator > 0:
            function_lines[i] = f"# {function_lines[i]}"
            in_decorator += line.count("(")
            in_decorator -= line.count(")")
        elif line.startswith("@"):
            decorator_name = line.split("@")[1].split(".")[0].split("(")[0]
            if decorator_name in cova_imports:
                function_lines[i] = f"# {function_lines[i]}"
                has_cova_decorator = True
                in_decorator += line.count("(")
                in_decorator -= line.count(")")

    return "\n".join(function_lines) if has_cova_decorator else function_string



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Create custom sentinels and defaults for Covalent"""

import os
from configparser import ConfigParser

import dask.system

prefix_separator = ":"

parameter_prefix = f"{prefix_separator}parameter{prefix_separator}"
electron_list_prefix = f"{prefix_separator}electron_list{prefix_separator}"
electron_dict_prefix = f"{prefix_separator}electron_dict{prefix_separator}"
subscript_prefix = f"{prefix_separator}subscripted{prefix_separator}"
generator_prefix = f"{prefix_separator}generated{prefix_separator}"
sublattice_prefix = f"{prefix_separator}sublattice{prefix_separator}"
attr_prefix = f"{prefix_separator}attribute{prefix_separator}"
arg_prefix = f"{prefix_separator}arg{prefix_separator}"

# Default configuration settings
_DEFAULT_CONFIG = {
    "sdk": {
        "log_dir": (
            os.environ.get("COVALENT_LOGDIR")
            or (os.environ.get("XDG_CACHE_HOME") or (os.environ["HOME"] + "/.cache"))
        )
        + "/covalent",
        "log_level": os.environ.get("LOGLEVEL", "WARNING").lower(),
        "enable_logging": os.environ.get("COVALENT_LOG_TO_FILE", "false").lower(),
        "executor_dir": os.environ.get("COVALENT_EXECUTOR_DIR")
        or (os.environ.get("XDG_CONFIG_DIR") or (os.environ["HOME"] + "/.config"))
        + "/covalent/executor_plugins",
    },
    "dispatcher": {
        "address": "localhost",
        "port": 48008,
        "cache_dir": (os.environ.get("XDG_CACHE_HOME") or (os.environ["HOME"] + "/.cache"))
        + "/covalent",
        "results_dir": os.environ.get("COVALENT_RESULTS_DIR", "results"),
        "log_dir": (os.environ.get("XDG_CACHE_HOME") or (os.environ["HOME"] + "/.cache"))
        + "/covalent",
        "db_path": (os.environ.get("XDG_DATA_HOME"))
        or (os.environ["HOME"] + "/.local/share") + "/covalent/dispatcher_db.sqlite",
    },
    "dask": {
        "cache_dir": (os.environ.get("XDG_CACHE_HOME") or (os.environ["HOME"] + "/.cache"))
        + "/covalent",
        "log_dir": (os.environ.get("XDG_CACHE_HOME") or (os.environ["HOME"] + "/.cache"))
        + "/covalent",
        "mem_per_worker": "auto",
        "threads_per_worker": 1,
        "num_workers": dask.system.CPU_COUNT,
    },
    "workflow_data": {
        "storage_type": "local",
        "base_dir": (os.environ.get("XDG_DATA_HOME"))
        or (os.environ["HOME"] + "/.local/share") + "/covalent/workflow_data",
    },
    "user_interface": {
        "address": "localhost",
        "port": 48008,
        "dev_port": 49009,
        "log_dir": (os.environ.get("XDG_CACHE_HOME") or (os.environ["HOME"] + "/.cache"))
        + "/covalent",
    },
}


def set_executor() -> dict:
    """Sets the executor based on whether Dask service is running.

    Returns:
        "dask" as the executor if Dask is running and "local" if Dask is not running.
    """
    config_dir = (
        os.environ.get("COVALENT_CONFIG_DIR")
        or (os.environ.get("XDG_CONFIG_DIR") or (os.environ["HOME"] + "/.config"))
    ) + "/covalent"
    config_parser = ConfigParser()
    config_file = f"{config_dir}/covalent.conf"
    config_parser.read(config_file)

    try:
        executor = (
            {"executor": "dask"}
            if config_parser["dask"]["scheduler_address"]
            else {"executor": "local"}
        )
    except KeyError:
        executor = {"executor": "local"}
    return executor


# Going forward we may only want to return the executor field of DEFAULT_CONSTRAINT_VALUES
# The rest of those parameters will now be in this dictionary
_DEFAULT_CONSTRAINT_VALUES = set_executor()
_DEFAULT_CONSTRAINT_VALUES["deps"] = {}
_DEFAULT_CONSTRAINT_VALUES["call_before"] = []
_DEFAULT_CONSTRAINT_VALUES["call_after"] = []
_DEFAULT_CONSTRAINT_VALUES["workflow_executor"] = _DEFAULT_CONSTRAINT_VALUES["executor"]

_DEFAULT_CONSTRAINTS_DEPRECATED = {
    "schedule": False,
    "num_cpu": 1,
    "cpu_feature_set": [],
    "num_gpu": 0,
    "gpu_type": "",
    "gpu_compute_capability": [],
    "memory": "1G",
    "executor": "local",
    "time_limit": "00-00:00:00",
    "budget": 0,
    "conda_env": "",
}

WAIT_EDGE_NAME = "!waiting_edge"



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Lattice route"""

import uuid
from random import randrange

from fastapi import APIRouter, HTTPException
from fastapi.responses import JSONResponse
from sqlalchemy.orm import Session

from covalent_ui.api.v1.data_layer.lattice_dal import Lattices
from covalent_ui.api.v1.database.config.db import engine
from covalent_ui.api.v1.models.lattices_model import (
    FileOutput,
    LatticeDetailResponse,
    LatticeExecutorResponse,
    LatticeFileResponse,
    LatticeWorkflowExecutorResponse,
)
from covalent_ui.api.v1.utils.file_handle import FileHandler

routes: APIRouter = APIRouter()


@routes.get("/{dispatch_id}", response_model=LatticeDetailResponse)
def get_lattice_details(dispatch_id: uuid.UUID):
    """Get lattice details

    Args:
        dispatch_id: To fetch lattice data with the provided dispatch id

    Returns:
        Returns the lattice data with the dispatch id provided
    """

    with Session(engine) as session:
        electron = Lattices(session)
        data = electron.get_lattices_id(dispatch_id)
        if data is not None:
            return LatticeDetailResponse(
                dispatch_id=data.dispatch_id,
                status=data.status,
                total_electrons=data.total_electrons,
                total_electrons_completed=data.total_electrons_completed,
                started_at=data.start_time,
                ended_at=data.end_time,
                directory=data.directory,
                runtime=data.runtime,
            )
        raise HTTPException(
            status_code=400,
            detail=[
                {
                    "loc": ["path", "dispatch_id"],
                    "msg": f"Dispatch ID {dispatch_id} does not exist",
                    "type": None,
                }
            ],
        )


@routes.get("/{dispatch_id}/details/{name}")
def get_lattice_files(dispatch_id: uuid.UUID, name: FileOutput):
    """Get lattice file data

    Args:
        dispatch_id: To fetch lattice data with the provided dispatch id
        name: To fetch specific file data for a lattice

    Returns:
        Returns the lattice file data with the dispatch id and file_module provided provided
    """
    with Session(engine) as session:
        electron = Lattices(session)
        lattice_data = electron.get_lattices_id_storage_file(dispatch_id)
        if lattice_data is not None:
            handler = FileHandler(lattice_data["directory"])
            if name == "result":
                response = handler.read_from_pickle(lattice_data["results_filename"])
                return LatticeFileResponse(data=str(response))
            if name == "inputs":
                response = handler.read_from_pickle(lattice_data["inputs_filename"])
                return LatticeFileResponse(data=response)
            elif name == "function_string":
                response = handler.read_from_text(lattice_data["function_string_filename"])
                return LatticeFileResponse(data=response)
            elif name == "executor":
                executor_name = lattice_data["executor"]
                executor_data = handler.read_from_pickle(lattice_data["executor_data_filename"])
                return LatticeExecutorResponse(
                    executor_name=executor_name, executor_details=executor_data
                )
            elif name == "workflow_executor":
                executor_name = lattice_data["workflow_executor"]
                executor_data = handler.read_from_pickle(
                    lattice_data["workflow_executor_data_filename"]
                )
                return LatticeWorkflowExecutorResponse(
                    workflow_executor_name=executor_name, workflow_executor_details=executor_data
                )
            elif name == "error":
                response = handler.read_from_text(lattice_data["error_filename"])
                return LatticeFileResponse(data=response)
            elif name == "function":
                response = handler.read_from_pickle(lattice_data["function_filename"])
                return LatticeFileResponse(data=response)
            elif name == "transport_graph":
                response = handler.read_from_pickle(lattice_data["transport_graph_filename"])
                return LatticeFileResponse(data=response)
            else:
                return LatticeFileResponse(data=None)
        else:
            raise HTTPException(
                status_code=400,
                detail=[
                    {
                        "loc": ["path", "dispatch_id"],
                        "msg": f"Dispatch ID {dispatch_id} does not exist",
                        "type": None,
                    }
                ],
            )



import enum


class Order(str, enum.Enum):
    BEFORE = "before"
    AFTER = "after"


class FileSchemes(str, enum.Enum):
    File = "file"
    S3 = "s3"
    Globus = "globus"
    HTTP = "http"
    HTTPS = "https"
    FTP = "ftp"


class FileTransferStrategyTypes(str, enum.Enum):
    Rsync = "rsync"
    HTTP = "http"
    S3 = "s3"
    FTP = "ftp"
    GLOBUS = "globus"


SchemeToStrategyMap = {
    "file": FileTransferStrategyTypes.Rsync,
    "http": FileTransferStrategyTypes.HTTP,
    "https": FileTransferStrategyTypes.HTTP,
    "s3": FileTransferStrategyTypes.S3,
    "ftp": FileTransferStrategyTypes.FTP,
    "globus": FileTransferStrategyTypes.GLOBUS,
}


class FtCallDepReturnValue(str, enum.Enum):
    TO = "to"
    FROM = "from"
    FROM_TO = "from_to"



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""
Class that defines the base executor template.
"""

import asyncio
import io
import os
import subprocess
import tempfile
from abc import ABC, abstractmethod
from contextlib import redirect_stderr, redirect_stdout
from pathlib import Path
from typing import Any, Callable, ContextManager, Dict, Iterable, List, Tuple

import aiofiles
import cloudpickle as pickle

from covalent._workflow.depscall import RESERVED_RETVAL_KEY__FILES

from .._shared_files import logger
from .._shared_files.context_managers import active_dispatch_info_manager
from .._shared_files.util_classes import DispatchInfo
from .._workflow.transport import TransportableObject

app_log = logger.app_log
log_stack_info = logger.log_stack_info


def wrapper_fn(
    function: TransportableObject,
    call_before: List[Tuple[TransportableObject, TransportableObject, TransportableObject]],
    call_after: List[Tuple[TransportableObject, TransportableObject, TransportableObject]],
    *args,
    **kwargs,
):
    """Wrapper for serialized callable.

    Execute preparatory shell commands before deserializing and
    running the callable. This is the actual function to be sent to
    the various executors.

    """

    cb_retvals = {}
    for tup in call_before:
        serialized_fn, serialized_args, serialized_kwargs, retval_key = tup
        cb_fn = serialized_fn.get_deserialized()
        cb_args = serialized_args.get_deserialized()
        cb_kwargs = serialized_kwargs.get_deserialized()
        retval = cb_fn(*cb_args, **cb_kwargs)

        # we always store cb_kwargs dict values as arrays to factor in non-unique values
        if retval_key and retval_key in cb_retvals:
            cb_retvals[retval_key].append(retval)
        elif retval_key:
            cb_retvals[retval_key] = [retval]

    # if cb_retvals key only contains one item this means it is a unique (non-repeated) retval key
    # so we only return the first element however if it is a 'files' kwarg we always return as a list
    cb_retvals = {
        key: value[0] if len(value) == 1 and key != RESERVED_RETVAL_KEY__FILES else value
        for key, value in cb_retvals.items()
    }

    fn = function.get_deserialized()

    new_args = [arg.get_deserialized() for arg in args]

    new_kwargs = {k: v.get_deserialized() for k, v in kwargs.items()}

    # Inject return values into kwargs
    for key, val in cb_retvals.items():
        new_kwargs[key] = val

    output = fn(*new_args, **new_kwargs)

    for tup in call_after:
        serialized_fn, serialized_args, serialized_kwargs, retval_key = tup
        ca_fn = serialized_fn.get_deserialized()
        ca_args = serialized_args.get_deserialized()
        ca_kwargs = serialized_kwargs.get_deserialized()
        ca_fn(*ca_args, **ca_kwargs)

    return TransportableObject(output)


class _AbstractBaseExecutor(ABC):
    """
    Private class that contains attributes and methods common to both
    BaseExecutor and BaseAsyncExecutor
    """

    def get_dispatch_context(self, dispatch_info: DispatchInfo) -> ContextManager[DispatchInfo]:
        """
        Start a context manager that will be used to
        access the dispatch info for the executor.

        Args:
            dispatch_info: The dispatch info to be used inside current context.

        Returns:
            A context manager object that handles the dispatch info.
        """

        return active_dispatch_info_manager.claim(dispatch_info)

    def short_name(self):
        module = self.__module__
        return self.__module__.split("/")[-1].split(".")[-1]

    def to_dict(self) -> dict:
        """Return a JSON-serializable dictionary representation of self"""
        return {
            "type": str(self.__class__),
            "short_name": self.short_name(),
            "attributes": self.__dict__.copy(),
        }

    def from_dict(self, object_dict: dict) -> "BaseExecutor":
        """Rehydrate a dictionary representation

        Args:
            object_dict: a dictionary representation returned by `to_dict`

        Returns:
            self

        Instance attributes will be overwritten.
        """
        if object_dict:
            self.__dict__ = object_dict["attributes"]
        return self


class BaseExecutor(_AbstractBaseExecutor):
    """
    Base executor class to be used for defining any executor
    plugin. Subclassing this class will allow you to define
    your own executor plugin which can be used in covalent.

    Note: When using a conda environment, it is assumed that
    covalent with all its dependencies are also installed in
    that environment.

    Attributes:
        log_stdout: The path to the file to be used for redirecting stdout.
        log_stderr: The path to the file to be used for redirecting stderr.
        conda_env: The name of the Conda environment to be used.
        cache_dir: The location used for cached files in the executor.
        current_env_on_conda_fail: If True, the current environment will be used
                                   if conda fails to activate specified env.
    """

    def __init__(
        self,
        log_stdout: str = "",
        log_stderr: str = "",
        conda_env: str = "",
        cache_dir: str = "",
        current_env_on_conda_fail: bool = False,
        *args,
        **kwargs,
    ) -> None:

        self.log_stdout = log_stdout
        self.log_stderr = log_stderr
        self.conda_env = conda_env
        self.cache_dir = cache_dir
        self.current_env_on_conda_fail = current_env_on_conda_fail
        self.current_env = ""

    def write_streams_to_file(
        self,
        stream_strings: Iterable[str],
        filepaths: Iterable[str],
        dispatch_id: str,
        results_dir: str,
    ) -> None:
        """
        Write the contents of stdout and stderr to respective files.

        Args:
            stream_strings: The stream_strings to be written to files.
            filepaths: The filepaths to be used for writing the streams.
            dispatch_id: The ID of the dispatch which initiated the request.
            results_dir: The location of the results directory.
        """

        for ss, filepath in zip(stream_strings, filepaths):
            if filepath:
                # If it is a relative path, attach to results dir
                if not Path(filepath).expanduser().is_absolute():
                    filepath = os.path.join(results_dir, dispatch_id, filepath)

                filename = Path(filepath)
                filename = filename.expanduser()
                filename.parent.mkdir(parents=True, exist_ok=True)
                filename.touch(exist_ok=True)

                with open(filepath, "a") as f:
                    f.write(ss)

    def execute(
        self,
        function: Callable,
        args: List,
        kwargs: Dict,
        dispatch_id: str,
        results_dir: str,
        node_id: int = -1,
    ) -> Any:
        """
        Execute the function with the given arguments.

        This calls the executor-specific `run()` method.

        Args:
            function: The input python function which will be executed and whose result
                      is ultimately returned by this function.
            args: List of positional arguments to be used by the function.
            kwargs: Dictionary of keyword arguments to be used by the function.
            dispatch_id: The unique identifier of the external lattice process which is
                         calling this function.
            results_dir: The location of the results directory.
            node_id: ID of the node in the transport graph which is using this executor.

        Returns:
            output: The result of the function execution.
        """

        dispatch_info = DispatchInfo(dispatch_id)
        fn_version = function.args[0].python_version

        task_metadata = {
            "dispatch_id": dispatch_id,
            "node_id": node_id,
            "results_dir": results_dir,
        }

        with self.get_dispatch_context(dispatch_info), redirect_stdout(
            io.StringIO()
        ) as stdout, redirect_stderr(io.StringIO()) as stderr:

            if self.conda_env != "":
                result = None

                result = self.execute_in_conda_env(
                    function,
                    fn_version,
                    args,
                    kwargs,
                    self.conda_env,
                    self.cache_dir,
                    node_id,
                )

            else:
                self.setup(task_metadata=task_metadata)
                result = self.run(function, args, kwargs, task_metadata)
                self.teardown(task_metadata=task_metadata)

        self.write_streams_to_file(
            (stdout.getvalue(), stderr.getvalue()),
            (self.log_stdout, self.log_stderr),
            dispatch_id,
            results_dir,
        )

        return (result, stdout.getvalue(), stderr.getvalue())

    def setup(self, task_metadata: Dict) -> Any:
        """Placeholder to run any executor specific tasks"""
        pass

    @abstractmethod
    def run(self, function: Callable, args: List, kwargs: Dict, task_metadata: Dict) -> Any:
        """Abstract method to run a function in the executor.

        Args:
            function: The function to run in the executor
            args: List of positional arguments to be used by the function
            kwargs: Dictionary of keyword arguments to be used by the function.
            task_metadata: Dictionary of metadata for the task. Current keys are
                          `dispatch_id` and `node_id`

        Returns:
            output: The result of the function execution
        """

        raise NotImplementedError

    def teardown(self, task_metadata: Dict) -> Any:
        """Placeholder to run nay executor specific cleanup/teardown actions"""
        pass

    def execute_in_conda_env(
        self,
        fn: Callable,
        fn_version: str,
        args: List,
        kwargs: Dict,
        conda_env: str,
        cache_dir: str,
        node_id: int,
    ) -> Tuple[bool, Any]:
        """
        Execute the function with the given arguments, in a Conda environment.

        Args:
            fn: The input python function which will be executed and whose result
                is ultimately returned by this function.
            fn_version: The python version the function was created with.
            args: List of positional arguments to be used by the function.
            kwargs: Dictionary of keyword arguments to be used by the function.
            conda_env: Name of a Conda environment in which to execute the task.
            cache_dir: The directory where temporary files and logs (if any) are stored.
            node_id: The integer identifier for the current node.

        Returns:
            output: The result of the function execution.
        """

        if not self.get_conda_path():
            return self._on_conda_env_fail(fn, args, kwargs, node_id)

        # Pickle the function
        temp_filename = ""
        with tempfile.NamedTemporaryFile(dir=cache_dir, delete=False) as f:
            pickle.dump(fn, f)
            temp_filename = f.name

        result_filename = os.path.join(cache_dir, f'result_{temp_filename.split("/")[-1]}')

        # Write a bash script to activate the environment
        shell_commands = "#!/bin/bash\n"

        # Add commands to initialize the Conda shell and activate the environment:
        conda_sh = os.path.join(
            os.path.dirname(self.conda_path), "..", "etc", "profile.d", "conda.sh"
        )
        conda_sh = os.environ.get("CONDA_SHELL", conda_sh)
        if os.path.exists(conda_sh):
            shell_commands += f"source {conda_sh}\n"
        else:
            message = "No Conda installation found on this compute node."
            app_log.warning(message)
            return self._on_conda_env_fail(fn, args, kwargs, node_id)

        shell_commands += f"conda activate {conda_env}\n"
        shell_commands += "retval=$?\n"
        shell_commands += "if [ $retval -ne 0 ]; then\n"
        shell_commands += (
            f'  echo "Conda environment {conda_env} is not present on this compute node."\n'
        )
        shell_commands += '  echo "Please create that environment (or use an existing environment) and try again."\n'
        shell_commands += "  exit 99\n"
        shell_commands += "fi\n\n"

        # Check Python version and give a warning if there is a mismatch:
        shell_commands += "py_version=`python -V | awk '{{print $2}}'`\n"
        shell_commands += f'if [[ "{fn_version}" != "$py_version" ]]; then\n'
        shell_commands += '  echo "Warning: Python version mismatch:"\n'
        shell_commands += f'  echo "Workflow version is {fn_version}. Conda environment version is $py_version."\n'
        shell_commands += "fi\n\n"

        shell_commands += "python - <<EOF\n"
        shell_commands += "import cloudpickle as pickle\n"
        shell_commands += "import os\n\n"

        # Add Python commands to run the pickled function:
        shell_commands += f'with open("{temp_filename}", "rb") as f:\n'
        shell_commands += "    fn = pickle.load(f)\n\n"

        shell_commands += f"result = fn(*{args}, **{kwargs})\n\n"

        shell_commands += f'with open("{result_filename}", "wb") as f:\n'
        shell_commands += "    pickle.dump(result, f)\n"
        shell_commands += "EOF\n"

        # Remove the temp file
        os.remove(temp_filename)

        # Run the script and unpickle the result
        with tempfile.NamedTemporaryFile(dir=cache_dir, mode="w") as f:
            f.write(shell_commands)
            f.flush()

            out = subprocess.run(["bash", f.name], capture_output=True, encoding="utf-8")

            if len(out.stdout) != 0:
                # These are print/log statements from the task.
                print(out.stdout)

            if out.returncode != 0:
                app_log.warning(out.stderr)
                return self._on_conda_env_fail(fn, args, kwargs, node_id)

        with open(result_filename, "rb") as f:
            result = pickle.load(f)

            message = f"Executed node {node_id} on Conda environment {self.conda_env}."
            app_log.debug(message)
            return result

    def _on_conda_env_fail(self, fn: Callable, args: List, kwargs: Dict, node_id: int):
        """

        Args:
            fn: The input python function which will be executed and
                whose result may be returned by this function.
            args: List of positional arguments to be used by the function.
            kwargs: Dictionary of keyword arguments to be used by the function.
            node_id: The integer identifier for the current node.

        Returns:
            output: The result of the function execution, if
                self.current_env_on_conda_fail == True, otherwise, return value is None.
        """

        result = None
        message = f"Failed to execute node {node_id} on Conda environment {self.conda_env}."
        if self.current_env_on_conda_fail:
            message += "\nExecuting on the current Conda environment."
            app_log.warning(message)
            result = fn(*args, **kwargs)

        else:
            app_log.error(message)
            raise RuntimeError

        return result

    def get_conda_envs(self) -> None:
        """
        Print a list of Conda environments detected on the system.

        Args:
            None

        Returns:
            None
        """

        self.conda_envs = []

        env_output = subprocess.run(
            ["conda", "env", "list"], capture_output=True, encoding="utf-8"
        )

        if len(env_output.stderr) > 0:
            message = f"Problem in listing Conda environments:\n{env_output.stderr}"
            app_log.warning(message)
            return

        for line in env_output.stdout.split("\n"):
            if not line.startswith("#"):
                row = line.split()
                if len(row) > 1:
                    if "*" in row:
                        self.current_env = row[0]
                    self.conda_envs.append(row[0])

        app_log.debug(f"Conda environments:\n{self.conda_envs}")

    def get_conda_path(self) -> bool:
        """
        Query the path where the conda executable can be found.

        Args:
            None

        Returns:
            found: True if Conda is found on the system.
        """

        self.conda_path = ""
        which_conda = subprocess.run(
            ["which", "conda"], capture_output=True, encoding="utf-8"
        ).stdout
        if which_conda == "":
            message = "No Conda installation found on this compute node."
            app_log.warning(message)
            return False
        self.conda_path = which_conda
        return True


class BaseAsyncExecutor(_AbstractBaseExecutor):
    def __init__(
        self,
        log_stdout: str = "",
        log_stderr: str = "",
        *args,
        **kwargs,
    ) -> None:

        self.log_stdout = log_stdout
        self.log_stderr = log_stderr

    async def write_streams_to_file(
        self,
        stream_strings: Iterable[str],
        filepaths: Iterable[str],
        dispatch_id: str,
        results_dir: str,
    ) -> None:

        """
        Write the contents of stdout and stderr to respective files.

        Args:
            stream_strings: The stream_strings to be written to files.
            filepaths: The filepaths to be used for writing the streams.
            dispatch_id: The ID of the dispatch which initiated the request.
            results_dir: The location of the results directory.

        This uses aiofiles to avoid blocking the event loop.
        """

        for ss, filepath in zip(stream_strings, filepaths):
            if filepath:
                # If it is a relative path, attach to results dir
                if not Path(filepath).expanduser().is_absolute():
                    filepath = os.path.join(results_dir, dispatch_id, filepath)

                filename = Path(filepath)
                filename = filename.expanduser()
                filename.parent.mkdir(parents=True, exist_ok=True)
                filename.touch(exist_ok=True)

                async with aiofiles.open(filepath, "a") as f:
                    await f.write(ss)

    async def execute(
        self,
        function: Callable,
        args: List,
        kwargs: Dict,
        dispatch_id: str,
        results_dir: str,
        node_id: int = -1,
    ) -> Any:

        task_metadata = {
            "dispatch_id": dispatch_id,
            "node_id": node_id,
            "results_dir": results_dir,
        }

        with redirect_stdout(io.StringIO()) as stdout, redirect_stderr(io.StringIO()) as stderr:
            await self.setup(task_metadata=task_metadata)
            result = await self.run(function, args, kwargs, task_metadata)
            await self.teardown(task_metadata=task_metadata)

        await self.write_streams_to_file(
            (stdout.getvalue(), stderr.getvalue()),
            (self.log_stdout, self.log_stderr),
            dispatch_id,
            results_dir,
        )

        return (result, stdout.getvalue(), stderr.getvalue())

    async def setup(self, task_metadata: Dict):
        """Executor specific setup method"""
        pass

    async def teardown(self, task_metadata: Dict):
        """Executor specific teardown method"""
        pass

    @abstractmethod
    async def run(self, function: Callable, args: List, kwargs: Dict, task_metadata: Dict) -> Any:
        """Abstract method to run a function in the executor in async-aware manner.

        Args:
            function: The function to run in the executor
            args: List of positional arguments to be used by the function
            kwargs: Dictionary of keyword arguments to be used by the function.
            task_metadata: Dictionary of metadata for the task. Current keys are
                          `dispatch_id` and `node_id`

        Returns:
            output: The result of the function execution
        """

        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import os
import shutil
import traceback
import uuid
from abc import ABC
from pathlib import Path
from typing import BinaryIO, Generator, List, Union

from .storagebackend import StorageBackend


def file_reader(filename: str):
    """Construct generator from file"""
    with open(filename, "rb") as f:
        yield from f


class LocalStorageBackend(StorageBackend):
    """Filesystem storage backend.

    Buckets = plain directories and object_names resolve to file
    paths. Custom metadata is not supported.

    Attributes:
        base_dir: root directory in which all buckets will be created
        bucket_name: current working bucket name (default: "default")

    """

    def __init__(self, base_dir: Path, bucket_name: str = "default"):
        self.base_dir = base_dir
        self.bucket_name = bucket_name

    def are_names_sane(self, bucket_name: str, object_name: str) -> bool:
        """Sanity check for bucket_name and object_name.

        Check that bucket name exists, resolves to a subdirectory of
        base_dir, and that bucket_name/object_name resolves to a
        location under bucket_name (e.g. bucket_name/../.. is
        forbidden)
        """

        p = self.base_dir / Path(bucket_name) / Path(object_name)

        try:
            abs_base = str(Path(self.base_dir).resolve(True))
            abs_bucket = str((abs_base / Path(bucket_name)).resolve(True))
        except FileNotFoundError:
            traceback.print_exception()
            traceback.print_stack()
            return False

        abs_p = str(p.resolve())

        if abs_bucket == abs_base or not abs_bucket.startswith(abs_base):
            return False

        if abs_p == abs_bucket or not abs_p.startswith(abs_bucket):
            return False

        return True

    def get(self, bucket_name: str, object_name: str) -> Union[Generator[bytes, None, None], None]:
        """Get object from storage.

        Args:
            bucket_name: name of the bucket
            object_name: name of the object

        Returns:
            A generator yielding a byte stream or None if an
            error occurred in retrieving the object

        """

        p = self.base_dir / Path(bucket_name) / Path(object_name)
        if not p.is_file():
            return None
        return file_reader(p)

    def put(
        self,
        data: BinaryIO,
        bucket_name: str,
        object_name: str,
        length: int,
        metadata: dict = None,
        overwrite: bool = False,
    ) -> (str, str):
        """Upload object to storage.

        Args:
            data: a binary file-like object
            bucket_name: name of the bucket
            object_name: name of the destination object
            length: number of bytes to upload
            metadata: not supported

        Returns:
            (bucket_name, object_name) if write succeeds and ("", "") otherwise

        """
        # TODO: Fix this to be less hacky.
        p = self.base_dir / Path(bucket_name) / Path(object_name)
        # This is a hack to get this working, but we need to think about if it's the right thing to do.
        try:
            os.makedirs(self.base_dir / Path(bucket_name))
        except FileExistsError:
            pass
        # Check that bucket name resolves to a subdirectory of base_dir
        # and bucket_name/object_name resolves to under bucket_name.
        try:
            abs_base = str(Path(self.base_dir).resolve(True))
            abs_bucket = str((abs_base / Path(bucket_name)).resolve(True))
        except FileNotFoundError:
            traceback.print_exception()
            traceback.print_stack()
            return ("", "")
        try:
            abs_p = str(p.resolve(True))
            object_name = f"{uuid.uuid4()}.pkl" if not overwrite else object_name
            p = self.base_dir / Path(bucket_name) / Path(object_name)
        except FileNotFoundError:
            pass
        abs_p = str(p.resolve())

        # TODO: improve error handling and logging
        if not abs_bucket.startswith(abs_base):
            return ("", "")

        if not abs_p.startswith(abs_bucket):
            return ("", "")

        if not p.parent.is_dir():
            p.parent.mkdir(parents=True)

        tmpdir = self.base_dir / Path(".tmp")
        if not tmpdir.is_dir():
            tmpdir.mkdir()

        tmppath = tmpdir / Path(object_name)

        try:
            # Ensure atomicity by writing to a temp file first and
            # then renaming it.

            with tmppath.open("wb") as tmp:
                shutil.copyfileobj(data, tmp)

            os.replace(tmppath, p)

            return (bucket_name, object_name)
        except:
            return ("", "")

    def delete(self, bucket_name: str, object_names: List[str]):
        deleted_objects = []
        failed = []
        for obj_name in object_names:
            if not self.are_names_sane(bucket_name, obj_name):
                continue

            p = self.base_dir / Path(bucket_name) / Path(obj_name)

            try:
                os.remove(p)
            except FileNotFoundError:
                failed.append(obj_name)
                pass
            except Exception:
                # e.g. permission errors
                failed.append(obj_name)
                continue

            deleted_objects.append(obj_name)

        return deleted_objects, failed



"""v11

Revision ID: 9b9d58f02985
Revises: b60c5ecdf927
Create Date: 2022-08-01 17:57:39.385604

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
# pragma: allowlist nextline secret
revision = "9b9d58f02985"
# pragma: allowlist nextline secret
down_revision = "b60c5ecdf927"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("electrons", sa.Column("executor", sa.Text(), nullable=True))
    op.add_column("electrons", sa.Column("executor_data_filename", sa.Text(), nullable=True))
    op.drop_column("electrons", "key_filename")
    op.drop_column("electrons", "attribute_name")
    op.drop_column("electrons", "executor_filename")
    op.add_column("lattices", sa.Column("executor", sa.Text(), nullable=True))
    op.add_column("lattices", sa.Column("executor_data_filename", sa.Text(), nullable=True))
    op.add_column("lattices", sa.Column("workflow_executor", sa.Text(), nullable=True))
    op.add_column(
        "lattices", sa.Column("workflow_executor_data_filename", sa.Text(), nullable=True)
    )
    op.add_column("lattices", sa.Column("named_args_filename", sa.Text(), nullable=True))
    op.add_column("lattices", sa.Column("named_kwargs_filename", sa.Text(), nullable=True))
    op.drop_column("lattices", "executor_filename")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("lattices", sa.Column("executor_filename", sa.TEXT(), nullable=True))
    op.drop_column("lattices", "named_kwargs_filename")
    op.drop_column("lattices", "named_args_filename")
    op.drop_column("lattices", "workflow_executor_data_filename")
    op.drop_column("lattices", "workflow_executor")
    op.drop_column("lattices", "executor_data_filename")
    op.drop_column("lattices", "executor")
    op.add_column("electrons", sa.Column("executor_filename", sa.TEXT(), nullable=True))
    op.add_column("electrons", sa.Column("attribute_name", sa.TEXT(), nullable=True))
    op.add_column("electrons", sa.Column("key_filename", sa.TEXT(), nullable=True))
    op.drop_column("electrons", "executor_data_filename")
    op.drop_column("electrons", "executor")
    # ### end Alembic commands ###



# from covalent._results_manager.results_manager import result_from
from covalent._results_manager import Result
from covalent._shared_files.defaults import (
    electron_dict_prefix,
    electron_list_prefix,
    parameter_prefix,
    prefix_separator,
    sublattice_prefix,
)
from covalent._workflow.transport import TransportableObject


def get_task_inputs(node_id: int, node_name: str, result_object: Result) -> dict:
    """
    Return the required inputs for a task execution.
    This makes sure that any node with child nodes isn't executed twice and fetches the
    result of parent node to use as input for the child node.

    Args:
        node_id: Node id of this task in the transport graph.
        node_name: Name of the node.
        result_object: Result object to be used to update and store execution related
                       info including the results.

    Returns:
        inputs: Input dictionary to be passed to the task containing args, kwargs,
                and any parent node execution results if present.
    """

    if node_name.startswith(electron_list_prefix):
        values = [
            result_object.lattice.transport_graph.get_node_value(parent, "output")
            for parent in result_object.lattice.transport_graph.get_dependencies(node_id)
        ]
        task_input = {"args": [], "kwargs": {"x": TransportableObject.make_transportable(values)}}
    elif node_name.startswith(electron_dict_prefix):
        values = {}
        for parent in result_object.lattice.transport_graph.get_dependencies(node_id):

            edge_data = result_object.lattice.transport_graph.get_edge_data(parent, node_id)

            value = result_object.lattice.transport_graph.get_node_value(parent, "output")
            for e_key, d in edge_data.items():
                key = d["edge_name"]
                values[key] = value

        task_input = {"args": [], "kwargs": {"x": TransportableObject.make_transportable(values)}}
    else:
        task_input = {"args": [], "kwargs": {}}

        for parent in result_object.lattice.transport_graph.get_dependencies(node_id):

            edge_data = result_object.lattice.transport_graph.get_edge_data(parent, node_id)
            value = result_object.lattice.transport_graph.get_node_value(parent, "output")

            for e_key, d in edge_data.items():
                if not d.get("wait_for"):
                    if d["param_type"] == "arg":
                        task_input["args"].append((value, d["arg_index"]))
                    elif d["param_type"] == "kwarg":
                        key = d["edge_name"]
                        task_input["kwargs"][key] = value

        sorted_args = sorted(task_input["args"], key=lambda x: x[1])
        task_input["args"] = [x[0] for x in sorted_args]

    return task_input



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Dispatch DataBase script."""

import copy
import json
import os
import sqlite3
from datetime import datetime
from typing import List, Tuple

import networkx as nx
import simplejson

import covalent.executor as covalent_executor
from covalent._data_store import DataStore
from covalent._results_manager.result import Result
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.util_classes import Status
from covalent._shared_files.utils import get_named_params

app_log = logger.app_log
log_stack_info = logger.log_stack_info
# DB Schema:
# TABLE dispatches
# * dispatch_id text primary key
# * result_dict text ; json-serialied dictionary representation of Result

# TODO: Move these to a common utils module


def extract_graph_node(node):
    # avoid mutating original node
    node = node.copy()

    # doc string
    f = node.get("function")
    if f is not None:
        node["doc"] = f.attrs["doc"]

    if "value" in node and node["value"] is not None:
        node["value"] = node["value"].object_string
    if "output" in node and node["output"] is not None:
        node["output"] = node["output"].object_string

    # metadata
    node["metadata"] = extract_metadata(node["metadata"])

    # prevent JSON encoding
    node["kwargs"] = encode_dict(node.get("kwargs"))

    # remove unused fields
    node.pop("function", None)
    node.pop("node_name", None)

    return node


def extract_metadata(metadata: dict):
    try:
        # avoid mutating original metadata
        metadata = copy.deepcopy(metadata)

        name = metadata["executor"]
        app_log.debug(f"Getting executor {name}")
        executor = covalent_executor._executor_manager.get_executor(name=name)

        if executor is not None:
            # extract attributes
            metadata["executor"] = encode_dict(executor.__dict__)
            if isinstance(name, str):
                metadata["executor_name"] = name
            else:
                metadata["executor_name"] = f"<{executor.__class__.__name__}>"

        metadata["deps"] = encode_dict(metadata["deps"])
        call_before = metadata["call_before"]
        call_after = metadata["call_after"]
        for i, dep in enumerate(call_before):
            call_before[i] = str(dep)

        for i, dep in enumerate(call_after):
            call_after[i] = str(dep)

        metadata["call_before"] = call_before
        metadata["call_after"] = call_after

    except (KeyError, AttributeError) as ex:
        app_log.error(f"Exception when trying to extract metadata: {ex}")

    return metadata


def encode_dict(d):
    """Avoid JSON encoding when python str() suffices"""
    if not isinstance(d, dict):
        return d
    return {k: str(v) for (k, v) in d.items()}


def extract_graph(graph):
    graph = nx.json_graph.node_link_data(graph)
    nodes = list(map(extract_graph_node, graph["nodes"]))
    return {
        "nodes": nodes,
        "links": graph["links"],
    }


def result_encoder(obj):
    if isinstance(obj, Status):
        return obj.STATUS
    if isinstance(obj, datetime):
        return obj.isoformat()
    return str(obj)


def encode_result(result_obj):
    lattice = result_obj.lattice

    result_string = result_obj.encoded_result.json
    if not result_string:
        result_string = result_obj.encoded_result.object_string

    named_args = {k: v.object_string for k, v in lattice.named_args.items()}
    named_kwargs = {k: v.object_string for k, v in lattice.named_kwargs.items()}
    result_dict = {
        "dispatch_id": result_obj.dispatch_id,
        "status": result_obj.status,
        "result": result_string,
        "start_time": result_obj.start_time,
        "end_time": result_obj.end_time,
        "results_dir": result_obj.results_dir,
        "error": result_obj.error,
        "lattice": {
            "function_string": lattice.workflow_function_string,
            "doc": lattice.__doc__,
            "name": lattice.__name__,
            "inputs": encode_dict({**named_args, **named_kwargs}),
            "metadata": extract_metadata(lattice.metadata),
        },
        "graph": extract_graph(result_obj.lattice.transport_graph._graph),
    }

    jsonified_result = simplejson.dumps(result_dict, default=result_encoder, ignore_nan=True)

    return jsonified_result


class DispatchDB:
    """
    Wrapper for the database of workflows.
    """

    def __init__(self, dbpath: str = None) -> None:
        if dbpath:
            self._dbpath = dbpath
        else:
            self._dbpath = get_config("user_interface.dispatch_db")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return False

    def _get_data_store(self, initialize_db: bool = False) -> DataStore:
        """Return the DataStore instance to write records."""

        return DataStore(db_URL=f"sqlite+pysqlite:///{self._dbpath}", initialize_db=initialize_db)



from abc import ABC, abstractmethod
from pathlib import Path

from covalent._file_transfer.enums import FtCallDepReturnValue
from covalent._file_transfer.file import File


class FileTransferStrategy(ABC):
    """
    Base FileTransferStrategy class that defines the interface for file transfer strategies exposing common methods for performing copy, download, and upload operations.

    """

    # move file (from) source (to) destination
    @abstractmethod
    def cp(self, from_file: File, to_file: File) -> None:
        raise NotImplementedError

    # download here implies 'from' is a remote source
    @abstractmethod
    def download(self, from_file: File, to_file: File) -> File:
        raise NotImplementedError

    # upload here implies 'to' is a remote source
    @abstractmethod
    def upload(self, from_file: File, to_file: File) -> None:
        raise NotImplementedError

    def pre_transfer_hook(
        self,
        from_file: File,
        to_file: File,
        return_value_type: FtCallDepReturnValue = FtCallDepReturnValue.FROM_TO,
    ) -> None:
        # Create any necessary temp files needed for file transfer operations
        from_filepath = from_file.filepath
        to_filepath = to_file.filepath
        is_from_temp_file = from_file.is_temp_file
        is_to_temp_file = to_file.is_temp_file

        # return value to be injected into kwargs of electron decorated fn as 'files'
        if return_value_type == FtCallDepReturnValue.TO:
            return_value = to_filepath
        elif return_value_type == FtCallDepReturnValue.FROM:
            return_value = from_filepath
        else:
            return_value = (from_filepath, to_filepath)

        def hook():
            if is_from_temp_file:
                from_path_obj = Path(from_filepath)
                from_path_obj.parent.mkdir(parents=True, exist_ok=True)
                from_path_obj.touch()
            if is_to_temp_file:
                to_path_obj = Path(to_filepath)
                to_path_obj.parent.mkdir(parents=True, exist_ok=True)
                to_path_obj.touch()
            return return_value

        return hook



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Lattice Data Layer"""

from datetime import timezone
from uuid import UUID

from sqlalchemy.orm import Session
from sqlalchemy.sql import func

from covalent_ui.api.v1.database.schema.electron import Electron
from covalent_ui.api.v1.database.schema.lattices import Lattice
from covalent_ui.api.v1.models.lattices_model import LatticeDetailResponse


class Lattices:
    """Lattice data access layer"""

    def __init__(self, db_con: Session) -> None:
        self.db_con = db_con

    def get_lattices_id(self, dispatch_id: UUID) -> LatticeDetailResponse:
        """
        Get lattices from dispatch id
        Args:
            dispatch_id: Refers to the dispatch_id in lattices table
        Return:
            Top most lattice with the given dispatch_id
            (i.e lattice with the same dispatch_id, but electron_id as null)
        """

        return (
            self.db_con.query(
                Lattice.dispatch_id,
                Lattice.status,
                Lattice.storage_path.label("directory"),
                Lattice.error_filename,
                func.datetime(Lattice.started_at, "localtime").label("start_time"),
                func.IFNULL(func.datetime(Lattice.completed_at, "localtime"), None).label(
                    "end_time"
                ),
                Lattice.electron_num.label("total_electrons"),
                Lattice.completed_electron_num.label("total_electrons_completed"),
                (
                    (
                        func.strftime(
                            "%s",
                            func.IFNULL(Lattice.completed_at, func.datetime.now(timezone.utc)),
                        )
                        - func.strftime("%s", Lattice.started_at)
                    )
                    * 1000
                ).label("runtime"),
                func.IFNULL(func.datetime(Lattice.updated_at, "localtime"), None).label(
                    "updated_at"
                ),
            )
            .join(Electron, Electron.parent_lattice_id == Lattice.id)
            .filter(Lattice.dispatch_id == str(dispatch_id), Lattice.is_active.is_not(False))
            .first()
        )

    def get_lattices_id_storage_file(self, dispatch_id: UUID):
        """
        Get storage file name
        Args:
            dispatch_id: Refers to the dispatch_id in lattices table
        Return:
            Top most lattice with the given dispatch_id along with file names
            (i.e lattice with the same dispatch_id, but electron_id as null)
        """

        return (
            self.db_con.query(
                Lattice.dispatch_id,
                Lattice.status,
                Lattice.storage_path.label("directory"),
                Lattice.error_filename,
                Lattice.function_string_filename,
                Lattice.executor,
                Lattice.executor_data_filename,
                Lattice.workflow_executor,
                Lattice.workflow_executor_data_filename,
                Lattice.error_filename,
                Lattice.inputs_filename,
                Lattice.results_filename,
                Lattice.storage_type,
                Lattice.function_filename,
                Lattice.transport_graph_filename,
                Lattice.started_at.label("started_at"),
                Lattice.completed_at.label("ended_at"),
                Lattice.electron_num.label("total_electrons"),
                Lattice.completed_electron_num.label("total_electrons_completed"),
            )
            .join(Electron, Electron.parent_lattice_id == Lattice.id)
            .filter(Lattice.dispatch_id == str(dispatch_id), Lattice.is_active.is_not(False))
            .first()
        )

    def get_lattice_id_by_dispatch_id(self, dispatch_id: UUID):
        """
        Get top lattice id from dispatch id
        Args:
            dispatch_id: UUID of dispatch
        Returns:
            Top most lattice id
        """
        data = (
            self.db_con.query(Lattice.id)
            .filter(Lattice.dispatch_id == str(dispatch_id), Lattice.electron_id.is_(None))
            .first()
        )
        return data[0]



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Summary Routes"""

from typing import Optional

from fastapi import APIRouter, Query
from pydantic import conint
from sqlalchemy.orm import Session

from covalent_ui.api.v1.data_layer.summary_dal import Summary
from covalent_ui.api.v1.database.config.db import engine
from covalent_ui.api.v1.models.dispatch_model import (
    DeleteDispatchesRequest,
    DeleteDispatchesResponse,
    DispatchDashBoardResponse,
    SortBy,
    SortDirection,
)

routes: APIRouter = APIRouter()


@routes.get("/list")
def get_all_dispatches(
    count: Optional[conint(gt=0, lt=100)] = Query(10),
    offset: Optional[int] = Query(0),
    sort_by: Optional[SortBy] = SortBy.RUNTIME,
    search: Optional[str] = "",
    sort_direction: Optional[SortDirection] = SortDirection.DESCENDING,
):
    """Get All Dispatches

    Args:
        req: Dispatch Summary Request

    Returns:
        List of Dispatch Summary
    """
    with Session(engine) as session:
        summary = Summary(session)
        return summary.get_summary(count, offset, sort_by, search, sort_direction)


@routes.get("/overview", response_model=DispatchDashBoardResponse)
def get_dashboard_details():
    """Overview of Dispatches

    Args:
        None

    Returns:
        An Overview of dispatches as object
    """
    with Session(engine) as session:
        summary = Summary(session)
        return summary.get_summary_overview()


@routes.post("/delete", response_model=DeleteDispatchesResponse)
def delete_dispatches(req: DeleteDispatchesRequest):
    """Delete one or more Dispatches

    Args:
        req: Dispatch Dispatch Request

    Returns:
        List of deleted dispatches if success
        List of deleted dispatches if fails
    """

    with Session(engine) as session:
        summary = Summary(session)
        return summary.delete_dispatches(req)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

from datetime import datetime, timezone
from sqlite3 import InterfaceError
from typing import List

from sqlalchemy import update
from sqlalchemy.orm import Session
from sqlalchemy.sql import desc, func, or_

from covalent_ui.api.v1.database.schema.electron import Electron
from covalent_ui.api.v1.database.schema.lattices import Lattice
from covalent_ui.api.v1.models.dispatch_model import (
    DeleteDispatchesRequest,
    DeleteDispatchesResponse,
    DispatchDashBoardResponse,
    DispatchResponse,
    SortDirection,
)


class Summary:
    """Summery data access layer"""

    def __init__(self, db_con: Session) -> None:
        self.db_con = db_con

    def get_summary(self, count, offset, sort_by, search, sort_direction) -> List[Lattice]:
        """
        Get summary of top most lattices
        Args:
            req.count: number of rows to be selected
            req.offset: number rows to be skipped
            req.sort_by: sort by field name(run_time, status, started, lattice)
            req.search: search by text
            req.direction: sort by direction ASE, DESC
        Return:
            List of top most Lattices and count
        """
        data = (
            self.db_con.query(
                Lattice.dispatch_id.label("dispatch_id"),
                Lattice.name.label("lattice_name"),
                (
                    (
                        func.strftime(
                            "%s",
                            func.IFNULL(Lattice.completed_at, func.datetime.now(timezone.utc)),
                        )
                        - func.strftime("%s", Lattice.started_at)
                    )
                    * 1000
                ).label("runtime"),
                Lattice.electron_num.label("total_electrons"),
                Lattice.completed_electron_num.label("total_electrons_completed"),
                func.datetime(Lattice.started_at, "localtime").label("started_at"),
                func.IFNULL(func.datetime(Lattice.completed_at, "localtime"), None).label(
                    "ended_at"
                ),
                Lattice.status.label("status"),
                func.datetime(Lattice.updated_at, "localtime").label("updated_at"),
            )
            .filter(
                or_(
                    Lattice.name.ilike(f"%{search}%"),
                    Lattice.dispatch_id.ilike(f"%{search}%"),
                ),
                Lattice.is_active.is_not(False),
                Lattice.electron_id.is_(None),
            )
            .order_by(
                desc(sort_by.value)
                if sort_direction == SortDirection.DESCENDING
                else sort_by.value
            )
            .offset(offset)
            .limit(count)
            .all()
        )

        counter = (
            self.db_con.query(func.count(Lattice.id))
            .filter(
                or_(
                    Lattice.name.ilike(f"%{search}%"),
                    Lattice.dispatch_id.ilike(f"%{search}%"),
                ),
                Lattice.is_active.is_not(False),
                Lattice.electron_id.is_(None),
            )
            .first()
        )
        return DispatchResponse(items=data, total_count=counter[0])

    def get_summary_overview(self) -> Lattice:
        """
        Get summary overview
        Args:
            None
        Return:
            Total jobs running,
            Total jobs done,
            Latest running task status,
            Total dispatcher duration
        """
        query1 = self.db_con.query(
            (func.count(Lattice.id))
            .filter(
                Lattice.status == "RUNNING",
                Lattice.is_active.is_not(False),
                Lattice.electron_id.is_(None),
            )
            .label("total_jobs_running")
        ).first()
        query2 = self.db_con.query(
            (func.count(Lattice.id))
            .filter(
                # Lattice.status == "COMPLETED",
                or_(
                    Lattice.status == "COMPLETED",
                    Lattice.status == "POSTPROCESSING",
                    Lattice.status == "POSTPROCESSING_FAILED",
                    Lattice.status == "PENDING_POSTPROCESSING",
                ),
                Lattice.is_active.is_not(False),
                Lattice.electron_id.is_(None),
            )
            .label("total_jobs_done")
        ).first()
        query3 = (
            self.db_con.query(Lattice.status)
            .filter(Lattice.is_active.is_not(False), Lattice.electron_id.is_(None))
            .order_by(Lattice.updated_at.desc())
            .first()
        )
        query4 = (
            self.db_con.query(
                (
                    func.sum(
                        func.strftime("%s", Lattice.completed_at)
                        - func.strftime("%s", Lattice.started_at)
                    )
                    * 1000
                ).label("run_time")
            )
            .filter(Lattice.is_active.is_not(False), Lattice.electron_id.is_(None))
            .first()
        )
        query5 = self.db_con.query(
            (func.count(Lattice.id))
            .filter(
                Lattice.status == "FAILED",
                Lattice.is_active.is_not(False),
                Lattice.electron_id.is_(None),
            )
            .label("total_failed")
        ).first()
        query6 = self.db_con.query(
            (func.count(Lattice.id))
            .filter(Lattice.is_active.is_not(False), Lattice.electron_id.is_(None))
            .label("total_jobs")
        ).first()
        if query4 is None:
            query4 = 0
        return DispatchDashBoardResponse(
            total_jobs_running=query1[0],
            total_jobs_completed=query2[0],
            latest_running_task_status=query3[0] if query3 is not None else None,
            total_dispatcher_duration=query4[0] if query4 is not None else 0,
            total_jobs_failed=query5[0],
            total_jobs=query6[0],
        )

    def delete_dispatches(self, req: DeleteDispatchesRequest) -> Lattice:
        """
        Delete dispatches
        Args:
            List[dispatch_id]
        Return:
            list of status(i.e whether given dispatch id is deleted successfully or failed)
        """
        success = []
        failure = []
        for dispatch_id in req.dispatches:
            try:
                lattice_id = (
                    self.db_con.query(Lattice.id)
                    .filter(
                        Lattice.dispatch_id == str(dispatch_id), Lattice.is_active.is_not(False)
                    )
                    .first()
                )
                if lattice_id is not None:
                    query1 = (
                        update(Electron)
                        .where(Electron.parent_lattice_id == lattice_id[0])
                        .values(
                            {
                                Electron.updated_at: datetime.now(timezone.utc),
                                Electron.is_active: False,
                            }
                        )
                    )
                    query2 = (
                        update(Lattice)
                        .where(Lattice.id == lattice_id[0])
                        .values(
                            {
                                Lattice.updated_at: datetime.now(timezone.utc),
                                Lattice.is_active: False,
                            }
                        )
                    )
                    self.db_con.execute(query1)
                    self.db_con.execute(query2)
                    self.db_con.commit()
                    success.append(dispatch_id)
                else:
                    failure.append(dispatch_id)
            except InterfaceError:
                failure.append(dispatch_id)
        if (len(failure) == 0 and len(success) == 0) or (len(failure) > 0 and len(success) == 0):
            message = "No dispatches were deleted"
        elif len(failure) > 0 and len(success) > 0:
            message = "Some of the dispatches could not be deleted"
        else:
            message = "Dispatch(es) have been deleted successfully!"
        return DeleteDispatchesResponse(
            success_items=success,
            failure_items=failure,
            message=message,
        )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Fastapi init"""

import socketio
from fastapi import FastAPI, HTTPException, Request
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse

from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent_ui.api.v1.routes import routes

# Config

WEBHOOK_PATH = "/api/webhook"
address = get_config("user_interface.address")
port = str(get_config("user_interface.dev_port"))
socket_port = str(get_config("user_interface.port"))
origins = [f"http://{address}:{port}"]
socket_origins = [f"http://{address}:{port}", f"http://{address}:{socket_port}"]

app_log = logger.app_log

app = FastAPI()
sio = socketio.AsyncServer(cors_allowed_origins=socket_origins, async_mode="asgi")


@sio.on("message")
async def chat_message(sid, data):
    await sio.emit("draw-request", "hi ")


app.include_router(routes.routes)

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request: Request, exc: RequestValidationError):
    return JSONResponse(
        status_code=422,
        content=jsonable_encoder({"detail": exc.errors()}),
    )


@app.exception_handler(HTTPException)
async def http_exception_handler(request, exc: HTTPException):
    return JSONResponse(
        status_code=400,
        content=jsonable_encoder({"detail": exc.detail}),
    )


@app.post(WEBHOOK_PATH)
async def handle_result_update(result_update: dict):
    await sio.emit("result-update", result_update)
    return {"ok": True}


@app.post("/api/draw")
async def handle_draw_request(draw_request: dict):
    await sio.emit("draw_request", draw_request)
    return {"ok": True}



import os
from pathlib import Path
from subprocess import PIPE, CalledProcessError, Popen
from typing import Optional

from covalent._file_transfer import File
from covalent._file_transfer.enums import FileSchemes, FileTransferStrategyTypes
from covalent._file_transfer.strategies.transfer_strategy_base import FileTransferStrategy


class Rsync(FileTransferStrategy):
    """
    Implements Base FileTransferStrategy class to use rsync to move files to and from remote or local filesystems. Rsync via ssh is used if one of the provided files is marked as remote.

    Attributes:
        user: (optional) Determine user to specify for remote host if using rsync with ssh
        host: (optional) Determine what host to connect to if using rsync with ssh
        private_key_path: (optional) Filepath for ssh private key to use if using rsync with ssh
    """

    def __init__(
        self,
        user: Optional[str] = "",
        host: Optional[str] = "",
        private_key_path: Optional[str] = None,
    ):

        self.user = user
        self.private_key_path = private_key_path
        self.host = host

        if self.private_key_path and not os.path.exists(self.private_key_path):
            raise FileNotFoundError(
                f"Provided private key ({self.private_key_path}) does not exist. Could not instantiate Rsync File Transfer Strategy. "
            )

    def get_rsync_ssh_cmd(
        self, local_file: File, remote_file: File, transfer_from_remote: bool = False
    ) -> str:
        local_filepath = local_file.filepath
        remote_filepath = remote_file.filepath
        args = ["rsync"]
        if self.private_key_path:
            args.append(f'-e "ssh -i {self.private_key_path}"')
        else:
            args.append("-e ssh")

        remote_source = f"{self.user}@{self.host}:{remote_filepath}"

        if transfer_from_remote:
            args.append(remote_source)
            args.append(local_filepath)
        else:
            args.append(local_filepath)
            args.append(remote_source)

        return " ".join(args)

    def get_rsync_cmd(
        self, from_file: File, to_file: File, transfer_from_remote: bool = False
    ) -> str:
        from_filepath = from_file.filepath
        to_filepath = to_file.filepath
        return f"rsync -a {from_filepath} {to_filepath}"

    def return_subprocess_callable(self, cmd) -> None:
        def callable():
            p = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
            output, error = p.communicate()
            if p.returncode != 0:
                raise CalledProcessError(p.returncode, f'"{cmd}" with error: {str(error)}')

        return callable

    # return callable to move files in the local file system
    def cp(self, from_file: File, to_file: File = File()) -> None:
        cmd = self.get_rsync_cmd(from_file, to_file)
        return self.return_subprocess_callable(cmd)

    # return callable to download here implies 'from' is a remote source
    def download(self, from_file: File, to_file: File = File()) -> File:
        cmd = self.get_rsync_ssh_cmd(to_file, from_file, transfer_from_remote=True)
        return self.return_subprocess_callable(cmd)

    # return callable to upload here implies 'to' is a remote source
    def upload(self, from_file: File, to_file: File) -> None:
        cmd = self.get_rsync_ssh_cmd(from_file, to_file, transfer_from_remote=False)
        return self.return_subprocess_callable(cmd)
